"""Unit test suite for docx.opc.pkgreader module."""

import pytest

from docx.opc.constants import CONTENT_TYPE as CT
from docx.opc.constants import RELATIONSHIP_TARGET_MODE as RTM
from docx.opc.packuri import PackURI
from docx.opc.phys_pkg import _ZipPkgReader
from docx.opc.pkgreader import (
    PackageReader,
    _ContentTypeMap,
    _SerializedPart,
    _SerializedRelationship,
    _SerializedRelationships,
)

from ..unitutil.mock import (
    ANY,
    Mock,
    call,
    class_mock,
    function_mock,
    initializer_mock,
    instance_mock,
    loose_mock,
    method_mock,
    patch,
)
from .unitdata.types import a_Default, a_Types, an_Override


class DescribePackageReader:
    def it_can_construct_from_pkg_file(
        self, _init_, PhysPkgReader_, from_xml, _srels_for, _load_serialized_parts
    ):
        phys_reader = PhysPkgReader_.return_value
        content_types = from_xml.return_value
        pkg_srels = _srels_for.return_value
        sparts = _load_serialized_parts.return_value
        pkg_file = Mock(name="pkg_file")

        pkg_reader = PackageReader.from_file(pkg_file)

        PhysPkgReader_.assert_called_once_with(pkg_file)
        from_xml.assert_called_once_with(phys_reader.content_types_xml)
        _srels_for.assert_called_once_with(phys_reader, "/")
        _load_serialized_parts.assert_called_once_with(phys_reader, pkg_srels, content_types)
        phys_reader.close.assert_called_once_with()
        _init_.assert_called_once_with(ANY, content_types, pkg_srels, sparts)
        assert isinstance(pkg_reader, PackageReader)

    def it_can_iterate_over_the_serialized_parts(self, iter_sparts_fixture):
        pkg_reader, expected_iter_spart_items = iter_sparts_fixture
        iter_spart_items = list(pkg_reader.iter_sparts())
        assert iter_spart_items == expected_iter_spart_items

    def it_can_iterate_over_all_the_srels(self):
        # mockery ----------------------
        pkg_srels = ["srel1", "srel2"]
        sparts = [
            Mock(name="spart1", partname="pn1", srels=["srel3", "srel4"]),
            Mock(name="spart2", partname="pn2", srels=["srel5", "srel6"]),
        ]
        pkg_reader = PackageReader(None, pkg_srels, sparts)
        # exercise ---------------------
        generated_tuples = list(pkg_reader.iter_srels())
        # verify -----------------------
        expected_tuples = [
            ("/", "srel1"),
            ("/", "srel2"),
            ("pn1", "srel3"),
            ("pn1", "srel4"),
            ("pn2", "srel5"),
            ("pn2", "srel6"),
        ]
        assert generated_tuples == expected_tuples

    def it_can_load_serialized_parts(self, _SerializedPart_, _walk_phys_parts):
        # test data --------------------
        test_data = (
            ("/part/name1.xml", "app/vnd.type_1", "reltype1", "<Part_1/>", "srels_1"),
            ("/part/name2.xml", "app/vnd.type_2", "reltype2", "<Part_2/>", "srels_2"),
        )
        iter_vals = [(t[0], t[2], t[3], t[4]) for t in test_data]
        content_types = {t[0]: t[1] for t in test_data}
        # mockery ----------------------
        phys_reader = Mock(name="phys_reader")
        pkg_srels = Mock(name="pkg_srels")
        _walk_phys_parts.return_value = iter_vals
        _SerializedPart_.side_effect = expected_sparts = (
            Mock(name="spart_1"),
            Mock(name="spart_2"),
        )
        # exercise ---------------------
        retval = PackageReader._load_serialized_parts(phys_reader, pkg_srels, content_types)
        # verify -----------------------
        expected_calls = [
            call("/part/name1.xml", "app/vnd.type_1", "<Part_1/>", "reltype1", "srels_1"),
            call("/part/name2.xml", "app/vnd.type_2", "<Part_2/>", "reltype2", "srels_2"),
        ]
        assert _SerializedPart_.call_args_list == expected_calls
        assert retval == expected_sparts

    def it_can_walk_phys_pkg_parts(self, _srels_for):
        # test data --------------------
        # +----------+       +--------+
        # | pkg_rels |-----> | part_1 |
        # +----------+       +--------+
        #      |               |    ^
        #      v               v    |
        #   external         +--------+     +--------+
        #                    | part_2 |---> | part_3 |
        #                    +--------+     +--------+
        partname_1, partname_2, partname_3 = (
            "/part/name1.xml",
            "/part/name2.xml",
            "/part/name3.xml",
        )
        part_1_blob, part_2_blob, part_3_blob = ("<Part_1/>", "<Part_2/>", "<Part_3/>")
        reltype1, reltype2, reltype3 = ("reltype1", "reltype2", "reltype3")
        srels = [
            Mock(name="rId1", is_external=True),
            Mock(
                name="rId2",
                is_external=False,
                reltype=reltype1,
                target_partname=partname_1,
            ),
            Mock(
                name="rId3",
                is_external=False,
                reltype=reltype2,
                target_partname=partname_2,
            ),
            Mock(
                name="rId4",
                is_external=False,
                reltype=reltype1,
                target_partname=partname_1,
            ),
            Mock(
                name="rId5",
                is_external=False,
                reltype=reltype3,
                target_partname=partname_3,
            ),
        ]
        pkg_srels = srels[:2]
        part_1_srels = srels[2:3]
        part_2_srels = srels[3:5]
        part_3_srels = []
        # mockery ----------------------
        phys_reader = Mock(name="phys_reader")
        _srels_for.side_effect = [part_1_srels, part_2_srels, part_3_srels]
        phys_reader.blob_for.side_effect = [part_1_blob, part_2_blob, part_3_blob]
        # exercise ---------------------
        generated_tuples = list(PackageReader._walk_phys_parts(phys_reader, pkg_srels))
        # verify -----------------------
        expected_tuples = [
            (partname_1, part_1_blob, reltype1, part_1_srels),
            (partname_2, part_2_blob, reltype2, part_2_srels),
            (partname_3, part_3_blob, reltype3, part_3_srels),
        ]
        assert generated_tuples == expected_tuples

    def it_can_retrieve_srels_for_a_source_uri(self, _SerializedRelationships_):
        # mockery ----------------------
        phys_reader = Mock(name="phys_reader")
        source_uri = Mock(name="source_uri")
        rels_xml = phys_reader.rels_xml_for.return_value
        load_from_xml = _SerializedRelationships_.load_from_xml
        srels = load_from_xml.return_value
        # exercise ---------------------
        retval = PackageReader._srels_for(phys_reader, source_uri)
        # verify -----------------------
        phys_reader.rels_xml_for.assert_called_once_with(source_uri)
        load_from_xml.assert_called_once_with(source_uri.baseURI, rels_xml)
        assert retval == srels

    # fixtures -------------------------------------------------------

    @pytest.fixture
    def blobs_(self, request):
        blob_ = loose_mock(request, spec=str, name="blob_")
        blob_2_ = loose_mock(request, spec=str, name="blob_2_")
        return blob_, blob_2_

    @pytest.fixture
    def content_types_(self, request):
        content_type_ = loose_mock(request, spec=str, name="content_type_")
        content_type_2_ = loose_mock(request, spec=str, name="content_type_2_")
        return content_type_, content_type_2_

    @pytest.fixture
    def from_xml(self, request):
        return method_mock(request, _ContentTypeMap, "from_xml", autospec=False)

    @pytest.fixture
    def _init_(self, request):
        return initializer_mock(request, PackageReader)

    @pytest.fixture
    def iter_sparts_fixture(self, sparts_, partnames_, content_types_, reltypes_, blobs_):
        pkg_reader = PackageReader(None, None, sparts_)
        expected_iter_spart_items = [
            (partnames_[0], content_types_[0], reltypes_[0], blobs_[0]),
            (partnames_[1], content_types_[1], reltypes_[1], blobs_[1]),
        ]
        return pkg_reader, expected_iter_spart_items

    @pytest.fixture
    def _load_serialized_parts(self, request):
        return method_mock(request, PackageReader, "_load_serialized_parts", autospec=False)

    @pytest.fixture
    def partnames_(self, request):
        partname_ = loose_mock(request, spec=str, name="partname_")
        partname_2_ = loose_mock(request, spec=str, name="partname_2_")
        return partname_, partname_2_

    @pytest.fixture
    def PhysPkgReader_(self):
        p = patch("docx.opc.pkgreader.PhysPkgReader", spec_set=_ZipPkgReader)
        yield p.start()
        p.stop()

    @pytest.fixture
    def reltypes_(self, request):
        reltype_ = instance_mock(request, str, name="reltype_")
        reltype_2_ = instance_mock(request, str, name="reltype_2")
        return reltype_, reltype_2_

    @pytest.fixture
    def _SerializedPart_(self, request):
        return class_mock(request, "docx.opc.pkgreader._SerializedPart")

    @pytest.fixture
    def _SerializedRelationships_(self, request):
        return class_mock(request, "docx.opc.pkgreader._SerializedRelationships")

    @pytest.fixture
    def sparts_(self, request, partnames_, content_types_, reltypes_, blobs_):
        sparts_ = []
        for idx in range(2):
            name = "spart_%s" % (("%d_" % (idx + 1)) if idx else "")
            spart_ = instance_mock(
                request,
                _SerializedPart,
                name=name,
                partname=partnames_[idx],
                content_type=content_types_[idx],
                reltype=reltypes_[idx],
                blob=blobs_[idx],
            )
            sparts_.append(spart_)
        return sparts_

    @pytest.fixture
    def _srels_for(self, request):
        return method_mock(request, PackageReader, "_srels_for", autospec=False)

    @pytest.fixture
    def _walk_phys_parts(self, request):
        return method_mock(request, PackageReader, "_walk_phys_parts", autospec=False)


class Describe_ContentTypeMap:
    def it_can_construct_from_ct_item_xml(self, from_xml_fixture):
        content_types_xml, expected_defaults, expected_overrides = from_xml_fixture
        ct_map = _ContentTypeMap.from_xml(content_types_xml)
        assert ct_map._defaults == expected_defaults
        assert ct_map._overrides == expected_overrides

    def it_matches_an_override_on_case_insensitive_partname(self, match_override_fixture):
        ct_map, partname, content_type = match_override_fixture
        assert ct_map[partname] == content_type

    def it_falls_back_to_case_insensitive_extension_default_match(self, match_default_fixture):
        ct_map, partname, content_type = match_default_fixture
        assert ct_map[partname] == content_type

    def it_should_raise_on_partname_not_found(self):
        ct_map = _ContentTypeMap()
        with pytest.raises(KeyError):
            ct_map[PackURI("/!blat/rhumba.1x&")]

    def it_should_raise_on_key_not_instance_of_PackURI(self):
        ct_map = _ContentTypeMap()
        ct_map._overrides = {PackURI("/part/name1.xml"): "app/vnd.type1"}
        with pytest.raises(KeyError):
            ct_map["/part/name1.xml"]

    # fixtures ---------------------------------------------

    @pytest.fixture
    def from_xml_fixture(self):
        entries = (
            ("Default", "xml", CT.XML),
            ("Default", "PNG", CT.PNG),
            ("Override", "/ppt/presentation.xml", CT.PML_PRESENTATION_MAIN),
        )
        content_types_xml = self._xml_from(entries)
        expected_defaults = {}
        expected_overrides = {}
        for entry in entries:
            if entry[0] == "Default":
                ext = entry[1].lower()
                content_type = entry[2]
                expected_defaults[ext] = content_type
            elif entry[0] == "Override":
                partname, content_type = entry[1:]
                expected_overrides[partname] = content_type
        return content_types_xml, expected_defaults, expected_overrides

    @pytest.fixture(
        params=[
            ("/foo/bar.xml", "xml", "application/xml"),
            ("/foo/bar.PNG", "png", "image/png"),
            ("/foo/bar.jpg", "JPG", "image/jpeg"),
        ]
    )
    def match_default_fixture(self, request):
        partname_str, ext, content_type = request.param
        partname = PackURI(partname_str)
        ct_map = _ContentTypeMap()
        ct_map._add_override(PackURI("/bar/foo.xyz"), "application/xyz")
        ct_map._add_default(ext, content_type)
        return ct_map, partname, content_type

    @pytest.fixture(
        params=[
            ("/foo/bar.xml", "/foo/bar.xml"),
            ("/foo/bar.xml", "/FOO/Bar.XML"),
            ("/FoO/bAr.XmL", "/foo/bar.xml"),
        ]
    )
    def match_override_fixture(self, request):
        partname_str, should_match_partname_str = request.param
        partname = PackURI(partname_str)
        should_match_partname = PackURI(should_match_partname_str)
        content_type = "appl/vnd-foobar"
        ct_map = _ContentTypeMap()
        ct_map._add_override(partname, content_type)
        return ct_map, should_match_partname, content_type

    def _xml_from(self, entries):
        """
        Return XML for a [Content_Types].xml based on items in `entries`.
        """
        types_bldr = a_Types().with_nsdecls()
        for entry in entries:
            if entry[0] == "Default":
                ext, content_type = entry[1:]
                default_bldr = a_Default()
                default_bldr.with_Extension(ext)
                default_bldr.with_ContentType(content_type)
                types_bldr.with_child(default_bldr)
            elif entry[0] == "Override":
                partname, content_type = entry[1:]
                override_bldr = an_Override()
                override_bldr.with_PartName(partname)
                override_bldr.with_ContentType(content_type)
                types_bldr.with_child(override_bldr)
        return types_bldr.xml()


class Describe_SerializedPart:
    def it_remembers_construction_values(self):
        # test data --------------------
        partname = "/part/name.xml"
        content_type = "app/vnd.type"
        reltype = "http://rel/type"
        blob = "<Part/>"
        srels = "srels proxy"
        # exercise ---------------------
        spart = _SerializedPart(partname, content_type, reltype, blob, srels)
        # verify -----------------------
        assert spart.partname == partname
        assert spart.content_type == content_type
        assert spart.reltype == reltype
        assert spart.blob == blob
        assert spart.srels == srels


class Describe_SerializedRelationship:
    def it_remembers_construction_values(self):
        # test data --------------------
        rel_elm = Mock(
            name="rel_elm",
            rId="rId9",
            reltype="ReLtYpE",
            target_ref="docProps/core.xml",
            target_mode=RTM.INTERNAL,
        )
        # exercise ---------------------
        srel = _SerializedRelationship("/", rel_elm)
        # verify -----------------------
        assert srel.rId == "rId9"
        assert srel.reltype == "ReLtYpE"
        assert srel.target_ref == "docProps/core.xml"
        assert srel.target_mode == RTM.INTERNAL

    def it_knows_when_it_is_external(self):
        cases = (RTM.INTERNAL, RTM.EXTERNAL, "FOOBAR")
        expected_values = (False, True, False)
        for target_mode, expected_value in zip(cases, expected_values):
            rel_elm = Mock(
                name="rel_elm",
                rId=None,
                reltype=None,
                target_ref=None,
                target_mode=target_mode,
            )
            srel = _SerializedRelationship(None, rel_elm)
            assert srel.is_external is expected_value

    def it_can_calculate_its_target_partname(self):
        # test data --------------------
        cases = (
            ("/", "docProps/core.xml", "/docProps/core.xml"),
            ("/ppt", "viewProps.xml", "/ppt/viewProps.xml"),
            (
                "/ppt/slides",
                "../slideLayouts/slideLayout1.xml",
                "/ppt/slideLayouts/slideLayout1.xml",
            ),
        )
        for baseURI, target_ref, expected_partname in cases:
            # setup --------------------
            rel_elm = Mock(
                name="rel_elm",
                rId=None,
                reltype=None,
                target_ref=target_ref,
                target_mode=RTM.INTERNAL,
            )
            # exercise -----------------
            srel = _SerializedRelationship(baseURI, rel_elm)
            # verify -------------------
            assert srel.target_partname == expected_partname

    def it_raises_on_target_partname_when_external(self):
        rel_elm = Mock(
            name="rel_elm",
            rId="rId9",
            reltype="ReLtYpE",
            target_ref="docProps/core.xml",
            target_mode=RTM.EXTERNAL,
        )
        srel = _SerializedRelationship("/", rel_elm)
        with pytest.raises(ValueError, match="target_partname attribute on Relat"):
            srel.target_partname


class Describe_SerializedRelationships:
    def it_can_load_from_xml(self, parse_xml_, _SerializedRelationship_):
        # mockery ----------------------
        baseURI, rels_item_xml, rel_elm_1, rel_elm_2 = (
            Mock(name="baseURI"),
            Mock(name="rels_item_xml"),
            Mock(name="rel_elm_1"),
            Mock(name="rel_elm_2"),
        )
        rels_elm = Mock(name="rels_elm", Relationship_lst=[rel_elm_1, rel_elm_2])
        parse_xml_.return_value = rels_elm
        # exercise ---------------------
        srels = _SerializedRelationships.load_from_xml(baseURI, rels_item_xml)
        # verify -----------------------
        expected_calls = [
            call(baseURI, rel_elm_1),
            call(baseURI, rel_elm_2),
        ]
        parse_xml_.assert_called_once_with(rels_item_xml)
        assert _SerializedRelationship_.call_args_list == expected_calls
        assert isinstance(srels, _SerializedRelationships)

    def it_should_be_iterable(self):
        srels = _SerializedRelationships()
        try:
            for x in srels:
                pass
        except TypeError:
            msg = "_SerializedRelationships object is not iterable"
            pytest.fail(msg)

    # fixtures ---------------------------------------------

    @pytest.fixture
    def parse_xml_(self, request):
        return function_mock(request, "docx.opc.pkgreader.parse_xml")

    @pytest.fixture
    def _SerializedRelationship_(self, request):
        return class_mock(request, "docx.opc.pkgreader._SerializedRelationship")



"""Step implementations for document settings-related features."""

from behave import given, then, when
from behave.runner import Context

from docx import Document
from docx.settings import Settings

from helpers import test_docx

# given ====================================================


@given("a document having a settings part")
def given_a_document_having_a_settings_part(context: Context):
    context.document = Document(test_docx("doc-word-default-blank"))


@given("a document having no settings part")
def given_a_document_having_no_settings_part(context: Context):
    context.document = Document(test_docx("set-no-settings-part"))


@given("a Settings object {with_or_without} odd and even page headers as settings")
def given_a_Settings_object_with_or_without_odd_and_even_hdrs(
    context: Context, with_or_without: str
):
    testfile_name = {"with": "doc-odd-even-hdrs", "without": "sct-section-props"}[with_or_without]
    context.settings = Document(test_docx(testfile_name)).settings


# when =====================================================


@when("I assign {bool_val} to settings.odd_and_even_pages_header_footer")
def when_I_assign_value_to_settings_odd_and_even_pages_header_footer(
    context: Context, bool_val: str
):
    context.settings.odd_and_even_pages_header_footer = eval(bool_val)


# then =====================================================


@then("document.settings is a Settings object")
def then_document_settings_is_a_Settings_object(context: Context):
    document = context.document
    assert type(document.settings) is Settings


@then("settings.odd_and_even_pages_header_footer is {bool_val}")
def then_settings_odd_and_even_pages_header_footer_is(context: Context, bool_val: str):
    actual = context.settings.odd_and_even_pages_header_footer
    expected = eval(bool_val)
    assert actual == expected, "settings.odd_and_even_pages_header_footer is %s" % actual



"""Step implementations for block content containers."""

from behave import given, then, when
from behave.runner import Context

from docx import Document
from docx.table import Table

from helpers import test_docx

# given ===================================================


@given("a _Cell object with paragraphs and tables")
def given_a_cell_with_paragraphs_and_tables(context: Context):
    context.cell = Document(test_docx("blk-paras-and-tables")).tables[1].rows[0].cells[0]


@given("a Document object with paragraphs and tables")
def given_a_document_with_paragraphs_and_tables(context: Context):
    context.document = Document(test_docx("blk-paras-and-tables"))


@given("a document containing a table")
def given_a_document_containing_a_table(context: Context):
    context.document = Document(test_docx("blk-containing-table"))


@given("a Footer object with paragraphs and tables")
def given_a_footer_with_paragraphs_and_tables(context: Context):
    context.footer = Document(test_docx("blk-paras-and-tables")).sections[0].footer


@given("a Header object with paragraphs and tables")
def given_a_header_with_paragraphs_and_tables(context: Context):
    context.header = Document(test_docx("blk-paras-and-tables")).sections[0].header


@given("a paragraph")
def given_a_paragraph(context: Context):
    context.document = Document()
    context.paragraph = context.document.add_paragraph()


# when ====================================================


@when("I add a paragraph")
def when_add_paragraph(context: Context):
    document = context.document
    context.p = document.add_paragraph()


@when("I add a table")
def when_add_table(context: Context):
    rows, cols = 2, 2
    context.document.add_table(rows, cols)


# then =====================================================


@then("cell.iter_inner_content() produces the block-items in document order")
def then_cell_iter_inner_content_produces_the_block_items(context: Context):
    actual = [type(item).__name__ for item in context.cell.iter_inner_content()]
    expected = ["Paragraph", "Table", "Paragraph"]
    assert actual == expected, f"expected: {expected}, got: {actual}"


@then("document.iter_inner_content() produces the block-items in document order")
def then_document_iter_inner_content_produces_the_block_items(context: Context):
    actual = [type(item).__name__ for item in context.document.iter_inner_content()]
    expected = ["Table", "Paragraph", "Table", "Paragraph", "Table", "Paragraph"]
    assert actual == expected, f"expected: {expected}, got: {actual}"


@then("footer.iter_inner_content() produces the block-items in document order")
def then_footer_iter_inner_content_produces_the_block_items(context: Context):
    actual = [type(item).__name__ for item in context.footer.iter_inner_content()]
    expected = ["Paragraph", "Table", "Paragraph"]
    assert actual == expected, f"expected: {expected}, got: {actual}"


@then("header.iter_inner_content() produces the block-items in document order")
def then_header_iter_inner_content_produces_the_block_items(context: Context):
    actual = [type(item).__name__ for item in context.header.iter_inner_content()]
    expected = ["Table", "Paragraph"]
    assert actual == expected, f"expected: {expected}, got: {actual}"


@then("I can access the table")
def then_can_access_table(context: Context):
    table = context.document.tables[-1]
    assert isinstance(table, Table)


@then("the new table appears in the document")
def then_new_table_appears_in_document(context: Context):
    table = context.document.tables[-1]
    assert isinstance(table, Table)



"""Step implementations for numbering-related features."""

from behave import given, then, when

from docx import Document

from helpers import test_docx

# given ===================================================


@given("a document having a numbering part")
def given_a_document_having_a_numbering_part(context):
    context.document = Document(test_docx("num-having-numbering-part"))


# when ====================================================


@when("I get the numbering part from the document")
def when_get_numbering_part_from_document(context):
    document = context.document
    context.numbering_part = document.part.numbering_part


# then =====================================================


@then("the numbering part has the expected numbering definitions")
def then_numbering_part_has_expected_numbering_definitions(context):
    numbering_part = context.numbering_part
    assert len(numbering_part.numbering_definitions) == 10



# pyright: reportPrivateUsage=false

"""Unit-test suite for `docx.oxml.comments` module."""

from __future__ import annotations

from typing import cast

import pytest

from docx.oxml.comments import CT_Comments

from ..unitutil.cxml import element


class DescribeCT_Comments:
    """Unit-test suite for `docx.oxml.comments.CT_Comments`."""

    @pytest.mark.parametrize(
        ("cxml", "expected_value"),
        [
            ("w:comments", 0),
            ("w:comments/(w:comment{w:id=1})", 2),
            ("w:comments/(w:comment{w:id=4},w:comment{w:id=2147483646})", 2147483647),
            ("w:comments/(w:comment{w:id=1},w:comment{w:id=2147483647})", 0),
            ("w:comments/(w:comment{w:id=1},w:comment{w:id=2},w:comment{w:id=3})", 4),
        ],
    )
    def it_finds_the_next_available_comment_id_to_help(self, cxml: str, expected_value: int):
        comments_elm = cast(CT_Comments, element(cxml))
        assert comments_elm._next_available_comment_id() == expected_value



"""Unit-test suite for `docx.oxml.section` module."""

from __future__ import annotations

from typing import cast

from docx.oxml.section import CT_HdrFtr
from docx.oxml.table import CT_Tbl
from docx.oxml.text.paragraph import CT_P

from ..unitutil.cxml import element


class DescribeCT_HdrFtr:
    """Unit-test suite for selected units of `docx.oxml.section.CT_HdrFtr`."""

    def it_knows_its_inner_content_block_item_elements(self):
        hdr = cast(CT_HdrFtr, element("w:hdr/(w:tbl,w:tbl,w:p)"))
        assert [type(e) for e in hdr.inner_content_elements] == [CT_Tbl, CT_Tbl, CT_P]



"""Test suite for the docx.oxml.text.run module."""

from typing import cast

import pytest

from docx.oxml.text.run import CT_R

from ...unitutil.cxml import element, xml


class DescribeCT_R:
    """Unit-test suite for the CT_R (run, <w:r>) element."""

    @pytest.mark.parametrize(
        ("initial_cxml", "text", "expected_cxml"),
        [
            ("w:r", "foobar", 'w:r/w:t"foobar"'),
            ("w:r", "foobar ", 'w:r/w:t{xml:space=preserve}"foobar "'),
            (
                "w:r/(w:rPr/w:rStyle{w:val=emphasis}, w:cr)",
                "foobar",
                'w:r/(w:rPr/w:rStyle{w:val=emphasis}, w:cr, w:t"foobar")',
            ),
        ],
    )
    def it_can_add_a_t_preserving_edge_whitespace(
        self, initial_cxml: str, text: str, expected_cxml: str
    ):
        r = cast(CT_R, element(initial_cxml))
        expected_xml = xml(expected_cxml)

        r.add_t(text)

        assert r.xml == expected_xml

    def it_can_assemble_the_text_in_the_run(self):
        cxml = 'w:r/(w:br,w:cr,w:noBreakHyphen,w:ptab,w:t"foobar",w:tab)'
        r = cast(CT_R, element(cxml))

        assert r.text == "\n\n-\tfoobar\t"



"""Unit-test suite for `docx.oxml.document` module."""

from __future__ import annotations

from typing import cast

from docx.oxml.document import CT_Body
from docx.oxml.table import CT_Tbl
from docx.oxml.text.paragraph import CT_P

from ..unitutil.cxml import element


class DescribeCT_Body:
    """Unit-test suite for selected units of `docx.oxml.document.CT_Body`."""

    def it_knows_its_inner_content_block_item_elements(self):
        body = cast(CT_Body, element("w:body/(w:tbl, w:p,w:p)"))
        assert [type(e) for e in body.inner_content_elements] == [CT_Tbl, CT_P, CT_P]



# pyright: reportPrivateUsage=false

"""Unit-test suite for the docx.text.pagebreak module."""

from typing import cast

import pytest

from docx import types as t
from docx.oxml.text.paragraph import CT_P
from docx.text.pagebreak import RenderedPageBreak

from ..unitutil.cxml import element, xml


class DescribeRenderedPageBreak:
    """Unit-test suite for the docx.text.pagebreak.RenderedPageBreak object."""

    def it_raises_on_preceding_fragment_when_page_break_is_not_first_in_paragrah(
        self, fake_parent: t.ProvidesStoryPart
    ):
        p_cxml = 'w:p/(w:r/(w:t"abc",w:lastRenderedPageBreak,w:lastRenderedPageBreak))'
        p = cast(CT_P, element(p_cxml))
        lrpb = p.lastRenderedPageBreaks[-1]
        page_break = RenderedPageBreak(lrpb, fake_parent)

        with pytest.raises(ValueError, match="only defined on first rendered page-br"):
            page_break.preceding_paragraph_fragment

    def it_produces_None_for_preceding_fragment_when_page_break_is_leading(
        self, fake_parent: t.ProvidesStoryPart
    ):
        """A page-break with no preceding content is "leading"."""
        p_cxml = 'w:p/(w:pPr/w:ind,w:r/(w:lastRenderedPageBreak,w:t"foo",w:t"bar"))'
        p = cast(CT_P, element(p_cxml))
        lrpb = p.lastRenderedPageBreaks[0]
        page_break = RenderedPageBreak(lrpb, fake_parent)

        preceding_fragment = page_break.preceding_paragraph_fragment

        assert preceding_fragment is None

    def it_can_split_off_the_preceding_paragraph_content_when_in_a_run(
        self, fake_parent: t.ProvidesStoryPart
    ):
        p_cxml = (
            "w:p/("
            "  w:pPr/w:ind"
            '  ,w:r/(w:t"foo",w:lastRenderedPageBreak,w:t"bar")'
            '  ,w:r/w:t"barfoo"'
            ")"
        )
        p = cast(CT_P, element(p_cxml))
        lrpb = p.lastRenderedPageBreaks[0]
        page_break = RenderedPageBreak(lrpb, fake_parent)

        preceding_fragment = page_break.preceding_paragraph_fragment

        expected_cxml = 'w:p/(w:pPr/w:ind,w:r/w:t"foo")'
        assert preceding_fragment is not None
        assert preceding_fragment._p.xml == xml(expected_cxml)

    def and_it_can_split_off_the_preceding_paragraph_content_when_in_a_hyperlink(
        self, fake_parent: t.ProvidesStoryPart
    ):
        p_cxml = (
            "w:p/("
            "  w:pPr/w:ind"
            '  ,w:hyperlink/w:r/(w:t"foo",w:lastRenderedPageBreak,w:t"bar")'
            '  ,w:r/w:t"barfoo"'
            ")"
        )
        p = cast(CT_P, element(p_cxml))
        lrpb = p.lastRenderedPageBreaks[0]
        page_break = RenderedPageBreak(lrpb, fake_parent)

        preceding_fragment = page_break.preceding_paragraph_fragment

        expected_cxml = 'w:p/(w:pPr/w:ind,w:hyperlink/w:r/(w:t"foo",w:t"bar"))'
        assert preceding_fragment is not None
        assert preceding_fragment._p.xml == xml(expected_cxml)

    def it_raises_on_following_fragment_when_page_break_is_not_first_in_paragrah(
        self, fake_parent: t.ProvidesStoryPart
    ):
        p_cxml = 'w:p/(w:r/(w:lastRenderedPageBreak,w:lastRenderedPageBreak,w:t"abc"))'
        p = cast(CT_P, element(p_cxml))
        lrpb = p.lastRenderedPageBreaks[-1]
        page_break = RenderedPageBreak(lrpb, fake_parent)

        with pytest.raises(ValueError, match="only defined on first rendered page-br"):
            page_break.following_paragraph_fragment

    def it_produces_None_for_following_fragment_when_page_break_is_trailing(
        self, fake_parent: t.ProvidesStoryPart
    ):
        """A page-break with no following content is "trailing"."""
        p_cxml = 'w:p/(w:pPr/w:ind,w:r/(w:t"foo",w:t"bar",w:lastRenderedPageBreak))'
        p = cast(CT_P, element(p_cxml))
        lrpb = p.lastRenderedPageBreaks[0]
        page_break = RenderedPageBreak(lrpb, fake_parent)

        following_fragment = page_break.following_paragraph_fragment

        assert following_fragment is None

    def it_can_split_off_the_following_paragraph_content_when_in_a_run(
        self, fake_parent: t.ProvidesStoryPart
    ):
        p_cxml = 'w:p/(w:pPr/w:ind,w:r/(w:t"foo",w:lastRenderedPageBreak,w:t"bar"),w:r/w:t"foo")'
        p = cast(CT_P, element(p_cxml))
        lrpb = p.lastRenderedPageBreaks[0]
        page_break = RenderedPageBreak(lrpb, fake_parent)

        following_fragment = page_break.following_paragraph_fragment

        expected_cxml = 'w:p/(w:pPr/w:ind,w:r/w:t"bar",w:r/w:t"foo")'
        assert following_fragment is not None
        assert following_fragment._p.xml == xml(expected_cxml)

    def and_it_can_split_off_the_following_paragraph_content_when_in_a_hyperlink(
        self, fake_parent: t.ProvidesStoryPart
    ):
        p_cxml = (
            "w:p/("
            "  w:pPr/w:ind"
            '  ,w:hyperlink/w:r/(w:t"foo",w:lastRenderedPageBreak,w:t"bar")'
            '  ,w:r/w:t"baz"'
            '  ,w:r/w:t"qux"'
            ")"
        )
        p = cast(CT_P, element(p_cxml))
        lrpb = p.lastRenderedPageBreaks[0]
        page_break = RenderedPageBreak(lrpb, fake_parent)

        following_fragment = page_break.following_paragraph_fragment

        expected_cxml = 'w:p/(w:pPr/w:ind,w:r/w:t"baz",w:r/w:t"qux")'

        assert following_fragment is not None
        assert following_fragment._p.xml == xml(expected_cxml)



"""Step implementations for document comments-related features."""

import datetime as dt

from behave import given, then, when
from behave.runner import Context

from docx import Document
from docx.comments import Comment, Comments
from docx.drawing import Drawing

from helpers import test_docx

# given ====================================================


@given("a Comment object")
def given_a_comment_object(context: Context):
    context.comment = Document(test_docx("comments-rich-para")).comments.get(0)


@given("a Comment object containing an embedded image")
def given_a_comment_object_containing_an_embedded_image(context: Context):
    context.comment = Document(test_docx("comments-rich-para")).comments.get(1)


@given("a Comments object with {count} comments")
def given_a_comments_object_with_count_comments(context: Context, count: str):
    testfile_name = {"0": "doc-default", "4": "comments-rich-para"}[count]
    context.comments = Document(test_docx(testfile_name)).comments


@given("a default Comment object")
def given_a_default_comment_object(context: Context):
    context.comment = Document(test_docx("comments-rich-para")).comments.add_comment()


@given("a document having a comments part")
def given_a_document_having_a_comments_part(context: Context):
    context.document = Document(test_docx("comments-rich-para"))


@given("a document having no comments part")
def given_a_document_having_no_comments_part(context: Context):
    context.document = Document(test_docx("doc-default"))


# when =====================================================


@when('I assign "{author}" to comment.author')
def when_I_assign_author_to_comment_author(context: Context, author: str):
    context.comment.author = author


@when("I assign comment = comments.add_comment()")
def when_I_assign_comment_eq_add_comment(context: Context):
    context.comment = context.comments.add_comment()


@when('I assign comment = comments.add_comment(author="John Doe", initials="JD")')
def when_I_assign_comment_eq_comments_add_comment_with_author_and_initials(context: Context):
    context.comment = context.comments.add_comment(author="John Doe", initials="JD")


@when('I assign comment = document.add_comment(runs, "A comment", "John Doe", "JD")')
def when_I_assign_comment_eq_document_add_comment(context: Context):
    runs = list(context.document.paragraphs[0].runs)
    context.comment = context.document.add_comment(
        runs=runs,
        text="A comment",
        author="John Doe",
        initials="JD",
    )


@when('I assign "{initials}" to comment.initials')
def when_I_assign_initials(context: Context, initials: str):
    context.comment.initials = initials


@when("I assign para_text = comment.paragraphs[0].text")
def when_I_assign_para_text(context: Context):
    context.para_text = context.comment.paragraphs[0].text


@when("I assign paragraph = comment.add_paragraph()")
def when_I_assign_default_add_paragraph(context: Context):
    context.paragraph = context.comment.add_paragraph()


@when("I assign paragraph = comment.add_paragraph(text, style)")
def when_I_assign_add_paragraph_with_text_and_style(context: Context):
    context.para_text = text = "Comment text"
    context.para_style = style = "Normal"
    context.paragraph = context.comment.add_paragraph(text, style)


@when("I assign run = paragraph.add_run()")
def when_I_assign_paragraph_add_run(context: Context):
    context.run = context.paragraph.add_run()


@when("I call comments.get(2)")
def when_I_call_comments_get_2(context: Context):
    context.comment = context.comments.get(2)


# then =====================================================


@then("comment is a Comment object")
def then_comment_is_a_Comment_object(context: Context):
    assert type(context.comment) is Comment


@then('comment.author == "{author}"')
def then_comment_author_eq_author(context: Context, author: str):
    actual = context.comment.author
    assert actual == author, f"expected author '{author}', got '{actual}'"


@then("comment.author is the author of the comment")
def then_comment_author_is_the_author_of_the_comment(context: Context):
    actual = context.comment.author
    assert actual == "Steve Canny", f"expected author 'Steve Canny', got '{actual}'"


@then("comment.comment_id == 0")
def then_comment_id_is_0(context: Context):
    assert context.comment.comment_id == 0


@then("comment.comment_id is the comment identifier")
def then_comment_comment_id_is_the_comment_identifier(context: Context):
    assert context.comment.comment_id == 0


@then("comment.initials is the initials of the comment author")
def then_comment_initials_is_the_initials_of_the_comment_author(context: Context):
    initials = context.comment.initials
    assert initials == "SJC", f"expected initials 'SJC', got '{initials}'"


@then('comment.initials == "{initials}"')
def then_comment_initials_eq_initials(context: Context, initials: str):
    actual = context.comment.initials
    assert actual == initials, f"expected initials '{initials}', got '{actual}'"


@then("comment.paragraphs[{idx}] == paragraph")
def then_comment_paragraphs_idx_eq_paragraph(context: Context, idx: str):
    actual = context.comment.paragraphs[int(idx)]._p
    expected = context.paragraph._p
    assert actual == expected, "paragraphs do not compare equal"


@then('comment.paragraphs[{idx}].style.name == "{style}"')
def then_comment_paragraphs_idx_style_name_eq_style(context: Context, idx: str, style: str):
    actual = context.comment.paragraphs[int(idx)]._p.style
    expected = style
    assert actual == expected, f"expected style name '{expected}', got '{actual}'"


@then('comment.text == "{text}"')
def then_comment_text_eq_text(context: Context, text: str):
    actual = context.comment.text
    expected = text
    assert actual == expected, f"expected text '{expected}', got '{actual}'"


@then("comment.timestamp is the date and time the comment was authored")
def then_comment_timestamp_is_the_date_and_time_the_comment_was_authored(context: Context):
    assert context.comment.timestamp == dt.datetime(2025, 6, 7, 11, 20, 0, tzinfo=dt.timezone.utc)


@then("comments.get({id}) == comment")
def then_comments_get_comment_id_eq_comment(context: Context, id: str):
    comment_id = int(id)
    comment = context.comments.get(comment_id)

    assert type(comment) is Comment, f"expected a Comment object, got {type(comment)}"
    assert comment.comment_id == comment_id, (
        f"expected comment_id '{comment_id}', got '{comment.comment_id}'"
    )


@then("document.comments is a Comments object")
def then_document_comments_is_a_Comments_object(context: Context):
    document = context.document
    assert type(document.comments) is Comments


@then("I can extract the image from the comment")
def then_I_can_extract_the_image_from_the_comment(context: Context):
    paragraph = context.comment.paragraphs[0]
    run = paragraph.runs[2]
    drawing = next(d for d in run.iter_inner_content() if isinstance(d, Drawing))
    assert drawing.has_picture

    image = drawing.image

    assert image.content_type == "image/jpeg", f"got {image.content_type}"
    assert image.filename == "image.jpg", f"got {image.filename}"
    assert image.sha1 == "1be010ea47803b00e140b852765cdf84f491da47", f"got {image.sha1}"


@then("iterating comments yields {count} Comment objects")
def then_iterating_comments_yields_count_comments(context: Context, count: str):
    comment_iter = iter(context.comments)

    comment = next(comment_iter)
    assert type(comment) is Comment, f"expected a Comment object, got {type(comment)}"

    remaining = list(comment_iter)
    assert len(remaining) == int(count) - 1, "iterating comments did not yield the expected count"


@then("len(comment.paragraphs) == {count}")
def then_len_comment_paragraphs_eq_count(context: Context, count: str):
    actual = len(context.comment.paragraphs)
    expected = int(count)
    assert actual == expected, f"expected len(comment.paragraphs) of {expected}, got {actual}"


@then("len(comments) == {count}")
def then_len_comments_eq_count(context: Context, count: str):
    actual = len(context.comments)
    expected = int(count)
    assert actual == expected, f"expected len(comments) of {expected}, got {actual}"


@then("para_text is the text of the first paragraph in the comment")
def then_para_text_is_the_text_of_the_first_paragraph_in_the_comment(context: Context):
    actual = context.para_text
    expected = "Text with hyperlink https://google.com embedded."
    assert actual == expected, f"expected para_text '{expected}', got '{actual}'"


@then("paragraph.style == style")
def then_paragraph_style_eq_known_style(context: Context):
    actual = context.paragraph.style.name
    expected = context.para_style
    assert actual == expected, f"expected paragraph.style '{expected}', got '{actual}'"


@then('paragraph.style == "{style}"')
def then_paragraph_style_eq_style(context: Context, style: str):
    actual = context.paragraph._p.style
    expected = style
    assert actual == expected, f"expected paragraph.style '{expected}', got '{actual}'"


@then("paragraph.text == text")
def then_paragraph_text_eq_known_text(context: Context):
    actual = context.paragraph.text
    expected = context.para_text
    assert actual == expected, f"expected paragraph.text '{expected}', got '{actual}'"


@then('paragraph.text == ""')
def then_paragraph_text_eq_text(context: Context):
    actual = context.paragraph.text
    expected = ""
    assert actual == expected, f"expected paragraph.text '{expected}', got '{actual}'"


@then("run.iter_inner_content() yields a single Picture drawing")
def then_run_iter_inner_content_yields_a_single_picture_drawing(context: Context):
    inner_content = list(context.run.iter_inner_content())

    assert len(inner_content) == 1, (
        f"expected a single inner content element, got {len(inner_content)}"
    )
    inner_content_item = inner_content[0]
    assert isinstance(inner_content_item, Drawing)
    assert inner_content_item.has_picture


@then("the result is a Comment object with id 2")
def then_the_result_is_a_comment_object_with_id_2(context: Context):
    comment = context.comment
    assert type(comment) is Comment, f"expected a Comment object, got {type(comment)}"
    assert comment.comment_id == 2, f"expected comment_id `2`, got '{comment.comment_id}'"



# pyright: reportPrivateUsage=false

"""Unit test suite for the `docx.drawing` module."""

from __future__ import annotations

from typing import cast

import pytest

from docx.drawing import Drawing
from docx.image.image import Image
from docx.oxml.drawing import CT_Drawing
from docx.parts.document import DocumentPart
from docx.parts.image import ImagePart

from .unitutil.cxml import element
from .unitutil.mock import FixtureRequest, Mock, instance_mock


class DescribeDrawing:
    """Unit-test suite for `docx.drawing.Drawing` objects."""

    @pytest.mark.parametrize(
        ("cxml", "expected_value"),
        [
            ("w:drawing/wp:inline/a:graphic/a:graphicData/pic:pic", True),
            ("w:drawing/wp:anchor/a:graphic/a:graphicData/pic:pic", True),
            ("w:drawing/wp:inline/a:graphic/a:graphicData/a:grpSp", False),
            ("w:drawing/wp:anchor/a:graphic/a:graphicData/a:chart", False),
        ],
    )
    def it_knows_when_it_contains_a_Picture(
        self, cxml: str, expected_value: bool, document_part_: Mock
    ):
        drawing = Drawing(cast(CT_Drawing, element(cxml)), document_part_)
        assert drawing.has_picture == expected_value

    def it_provides_access_to_the_image_in_a_Picture_drawing(
        self, document_part_: Mock, image_part_: Mock, image_: Mock
    ):
        image_part_.image = image_
        document_part_.part.related_parts = {"rId1": image_part_}
        cxml = (
            "w:drawing/wp:inline/a:graphic/a:graphicData/pic:pic/pic:blipFill/a:blip{r:embed=rId1}"
        )
        drawing = Drawing(cast(CT_Drawing, element(cxml)), document_part_)

        image = drawing.image

        assert image is image_

    def but_it_raises_when_the_drawing_does_not_contain_a_Picture(self, document_part_: Mock):
        drawing = Drawing(
            cast(CT_Drawing, element("w:drawing/wp:inline/a:graphic/a:graphicData/a:grpSp")),
            document_part_,
        )

        with pytest.raises(ValueError, match="drawing does not contain a picture"):
            drawing.image

    # -- fixtures --------------------------------------------------------------------------------

    @pytest.fixture
    def document_part_(self, request: FixtureRequest):
        return instance_mock(request, DocumentPart)

    @pytest.fixture
    def image_(self, request: FixtureRequest):
        return instance_mock(request, Image)

    @pytest.fixture
    def image_part_(self, request: FixtureRequest):
        return instance_mock(request, ImagePart)



# pyright: reportPrivateUsage=false

"""Test suite for the docx.text.run module."""

from __future__ import annotations

from typing import Any, List, cast

import pytest

from docx import types as t
from docx.enum.style import WD_STYLE_TYPE
from docx.enum.text import WD_BREAK, WD_UNDERLINE
from docx.oxml.text.paragraph import CT_P
from docx.oxml.text.run import CT_R
from docx.parts.document import DocumentPart
from docx.shape import InlineShape
from docx.text.font import Font
from docx.text.paragraph import Paragraph
from docx.text.run import Run

from ..unitutil.cxml import element, xml
from ..unitutil.mock import FixtureRequest, Mock, class_mock, instance_mock, property_mock


class DescribeRun:
    """Unit-test suite for `docx.text.run.Run`."""

    @pytest.mark.parametrize(
        ("r_cxml", "bool_prop_name", "expected_value"),
        [
            ("w:r/w:rPr", "bold", None),
            ("w:r/w:rPr/w:b", "bold", True),
            ("w:r/w:rPr/w:b{w:val=on}", "bold", True),
            ("w:r/w:rPr/w:b{w:val=off}", "bold", False),
            ("w:r/w:rPr/w:b{w:val=1}", "bold", True),
            ("w:r/w:rPr/w:i{w:val=0}", "italic", False),
        ],
    )
    def it_knows_its_bool_prop_states(
        self, r_cxml: str, bool_prop_name: str, expected_value: bool | None, paragraph_: Mock
    ):
        run = Run(cast(CT_R, element(r_cxml)), paragraph_)
        assert getattr(run, bool_prop_name) == expected_value

    @pytest.mark.parametrize(
        ("initial_r_cxml", "bool_prop_name", "value", "expected_cxml"),
        [
            # -- nothing to True, False, and None ---------------------------
            ("w:r", "bold", True, "w:r/w:rPr/w:b"),
            ("w:r", "bold", False, "w:r/w:rPr/w:b{w:val=0}"),
            ("w:r", "italic", None, "w:r/w:rPr"),
            # -- default to True, False, and None ---------------------------
            ("w:r/w:rPr/w:b", "bold", True, "w:r/w:rPr/w:b"),
            ("w:r/w:rPr/w:b", "bold", False, "w:r/w:rPr/w:b{w:val=0}"),
            ("w:r/w:rPr/w:i", "italic", None, "w:r/w:rPr"),
            # -- True to True, False, and None ------------------------------
            ("w:r/w:rPr/w:b{w:val=on}", "bold", True, "w:r/w:rPr/w:b"),
            ("w:r/w:rPr/w:b{w:val=1}", "bold", False, "w:r/w:rPr/w:b{w:val=0}"),
            ("w:r/w:rPr/w:b{w:val=1}", "bold", None, "w:r/w:rPr"),
            # -- False to True, False, and None -----------------------------
            ("w:r/w:rPr/w:i{w:val=false}", "italic", True, "w:r/w:rPr/w:i"),
            ("w:r/w:rPr/w:i{w:val=0}", "italic", False, "w:r/w:rPr/w:i{w:val=0}"),
            ("w:r/w:rPr/w:i{w:val=off}", "italic", None, "w:r/w:rPr"),
        ],
    )
    def it_can_change_its_bool_prop_settings(
        self,
        initial_r_cxml: str,
        bool_prop_name: str,
        value: bool | None,
        expected_cxml: str,
        paragraph_: Mock,
    ):
        run = Run(cast(CT_R, element(initial_r_cxml)), paragraph_)

        setattr(run, bool_prop_name, value)

        assert run._r.xml == xml(expected_cxml)

    @pytest.mark.parametrize(
        ("r_cxml", "expected_value"),
        [
            ("w:r", False),
            ('w:r/w:t"foobar"', False),
            ('w:r/(w:t"abc", w:lastRenderedPageBreak, w:t"def")', True),
            ("w:r/(w:lastRenderedPageBreak, w:lastRenderedPageBreak)", True),
        ],
    )
    def it_knows_whether_it_contains_a_page_break(
        self, r_cxml: str, expected_value: bool, paragraph_: Mock
    ):
        run = Run(cast(CT_R, element(r_cxml)), paragraph_)
        assert run.contains_page_break == expected_value

    @pytest.mark.parametrize(
        ("r_cxml", "expected"),
        [
            # -- no content produces an empty iterator --
            ("w:r", []),
            # -- contiguous text content is condensed into a single str --
            ('w:r/(w:t"foo",w:cr,w:t"bar")', ["str"]),
            # -- page-breaks are a form of inner-content --
            (
                'w:r/(w:t"abc",w:br,w:lastRenderedPageBreak,w:noBreakHyphen,w:t"def")',
                ["str", "RenderedPageBreak", "str"],
            ),
            # -- as are drawings --
            (
                'w:r/(w:t"abc", w:lastRenderedPageBreak, w:drawing)',
                ["str", "RenderedPageBreak", "Drawing"],
            ),
        ],
    )
    def it_can_iterate_its_inner_content_items(
        self, r_cxml: str, expected: List[str], fake_parent: t.ProvidesStoryPart
    ):
        r = cast(CT_R, element(r_cxml))
        run = Run(r, fake_parent)

        inner_content = run.iter_inner_content()

        actual = [type(item).__name__ for item in inner_content]
        assert actual == expected, f"expected: {expected}, got: {actual}"

    def it_can_mark_a_comment_reference_range(self, paragraph_: Mock):
        p = cast(CT_P, element('w:p/w:r/w:t"referenced text"'))
        run = last_run = Run(p.r_lst[0], paragraph_)

        run.mark_comment_range(last_run, comment_id=42)

        assert p.xml == xml(
            'w:p/(w:commentRangeStart{w:id=42},w:r/w:t"referenced text"'
            ",w:commentRangeEnd{w:id=42}"
            ",w:r/(w:rPr/w:rStyle{w:val=CommentReference},w:commentReference{w:id=42}))"
        )

    def it_knows_its_character_style(
        self, part_prop_: Mock, document_part_: Mock, paragraph_: Mock
    ):
        style_ = document_part_.get_style.return_value
        part_prop_.return_value = document_part_
        style_id = "Barfoo"
        run = Run(cast(CT_R, element(f"w:r/w:rPr/w:rStyle{{w:val={style_id}}}")), paragraph_)

        style = run.style

        document_part_.get_style.assert_called_once_with(style_id, WD_STYLE_TYPE.CHARACTER)
        assert style is style_

    @pytest.mark.parametrize(
        ("r_cxml", "value", "style_id", "expected_cxml"),
        [
            ("w:r", "Foo Font", "FooFont", "w:r/w:rPr/w:rStyle{w:val=FooFont}"),
            ("w:r/w:rPr", "Foo Font", "FooFont", "w:r/w:rPr/w:rStyle{w:val=FooFont}"),
            (
                "w:r/w:rPr/w:rStyle{w:val=FooFont}",
                "Bar Font",
                "BarFont",
                "w:r/w:rPr/w:rStyle{w:val=BarFont}",
            ),
            ("w:r/w:rPr/w:rStyle{w:val=FooFont}", None, None, "w:r/w:rPr"),
            ("w:r", None, None, "w:r/w:rPr"),
        ],
    )
    def it_can_change_its_character_style(
        self,
        r_cxml: str,
        value: str | None,
        style_id: str | None,
        expected_cxml: str,
        part_prop_: Mock,
        paragraph_: Mock,
    ):
        part_ = part_prop_.return_value
        part_.get_style_id.return_value = style_id
        run = Run(cast(CT_R, element(r_cxml)), paragraph_)

        run.style = value

        part_.get_style_id.assert_called_once_with(value, WD_STYLE_TYPE.CHARACTER)
        assert run._r.xml == xml(expected_cxml)

    @pytest.mark.parametrize(
        ("r_cxml", "expected_value"),
        [
            ("w:r", None),
            ("w:r/w:rPr/w:u", None),
            ("w:r/w:rPr/w:u{w:val=single}", True),
            ("w:r/w:rPr/w:u{w:val=none}", False),
            ("w:r/w:rPr/w:u{w:val=double}", WD_UNDERLINE.DOUBLE),
            ("w:r/w:rPr/w:u{w:val=wave}", WD_UNDERLINE.WAVY),
        ],
    )
    def it_knows_its_underline_type(
        self, r_cxml: str, expected_value: bool | WD_UNDERLINE | None, paragraph_: Mock
    ):
        run = Run(cast(CT_R, element(r_cxml)), paragraph_)
        assert run.underline is expected_value

    @pytest.mark.parametrize(
        ("initial_r_cxml", "new_underline", "expected_cxml"),
        [
            ("w:r", True, "w:r/w:rPr/w:u{w:val=single}"),
            ("w:r", False, "w:r/w:rPr/w:u{w:val=none}"),
            ("w:r", None, "w:r/w:rPr"),
            ("w:r", WD_UNDERLINE.SINGLE, "w:r/w:rPr/w:u{w:val=single}"),
            ("w:r", WD_UNDERLINE.THICK, "w:r/w:rPr/w:u{w:val=thick}"),
            ("w:r/w:rPr/w:u{w:val=single}", True, "w:r/w:rPr/w:u{w:val=single}"),
            ("w:r/w:rPr/w:u{w:val=single}", False, "w:r/w:rPr/w:u{w:val=none}"),
            ("w:r/w:rPr/w:u{w:val=single}", None, "w:r/w:rPr"),
            (
                "w:r/w:rPr/w:u{w:val=single}",
                WD_UNDERLINE.SINGLE,
                "w:r/w:rPr/w:u{w:val=single}",
            ),
            (
                "w:r/w:rPr/w:u{w:val=single}",
                WD_UNDERLINE.DOTTED,
                "w:r/w:rPr/w:u{w:val=dotted}",
            ),
        ],
    )
    def it_can_change_its_underline_type(
        self,
        initial_r_cxml: str,
        new_underline: bool | WD_UNDERLINE | None,
        expected_cxml: str,
        paragraph_: Mock,
    ):
        run = Run(cast(CT_R, element(initial_r_cxml)), paragraph_)

        run.underline = new_underline

        assert run._r.xml == xml(expected_cxml)

    @pytest.mark.parametrize("invalid_value", ["foobar", 42, "single"])
    def it_raises_on_assign_invalid_underline_value(self, invalid_value: Any, paragraph_: Mock):
        run = Run(cast(CT_R, element("w:r/w:rPr")), paragraph_)
        with pytest.raises(ValueError, match=" is not a valid WD_UNDERLINE"):
            run.underline = invalid_value

    def it_provides_access_to_its_font(self, Font_: Mock, font_: Mock, paragraph_: Mock):
        Font_.return_value = font_
        run = Run(cast(CT_R, element("w:r")), paragraph_)

        font = run.font

        Font_.assert_called_once_with(run._element)
        assert font is font_

    @pytest.mark.parametrize(
        ("r_cxml", "new_text", "expected_cxml"),
        [
            ("w:r", "foo", 'w:r/w:t"foo"'),
            ('w:r/w:t"foo"', "bar", 'w:r/(w:t"foo", w:t"bar")'),
            ("w:r", "fo ", 'w:r/w:t{xml:space=preserve}"fo "'),
            ("w:r", "f o", 'w:r/w:t"f o"'),
        ],
    )
    def it_can_add_text(
        self, r_cxml: str, new_text: str, expected_cxml: str, Text_: Mock, paragraph_: Mock
    ):
        run = Run(cast(CT_R, element(r_cxml)), paragraph_)

        text = run.add_text(new_text)

        assert run._r.xml == xml(expected_cxml)
        assert text is Text_.return_value

    @pytest.mark.parametrize(
        ("break_type", "expected_cxml"),
        [
            (WD_BREAK.LINE, "w:r/w:br"),
            (WD_BREAK.PAGE, "w:r/w:br{w:type=page}"),
            (WD_BREAK.COLUMN, "w:r/w:br{w:type=column}"),
            (WD_BREAK.LINE_CLEAR_LEFT, "w:r/w:br{w:clear=left}"),
            (WD_BREAK.LINE_CLEAR_RIGHT, "w:r/w:br{w:clear=right}"),
            (WD_BREAK.LINE_CLEAR_ALL, "w:r/w:br{w:clear=all}"),
        ],
    )
    def it_can_add_a_break(self, break_type: WD_BREAK, expected_cxml: str, paragraph_: Mock):
        run = Run(cast(CT_R, element("w:r")), paragraph_)

        run.add_break(break_type)

        assert run._r.xml == xml(expected_cxml)

    @pytest.mark.parametrize(
        ("r_cxml", "expected_cxml"), [('w:r/w:t"foo"', 'w:r/(w:t"foo", w:tab)')]
    )
    def it_can_add_a_tab(self, r_cxml: str, expected_cxml: str, paragraph_: Mock):
        run = Run(cast(CT_R, element(r_cxml)), paragraph_)

        run.add_tab()

        assert run._r.xml == xml(expected_cxml)

    def it_can_add_a_picture(
        self,
        part_prop_: Mock,
        document_part_: Mock,
        InlineShape_: Mock,
        picture_: Mock,
        paragraph_: Mock,
    ):
        part_prop_.return_value = document_part_
        run = Run(cast(CT_R, element("w:r/wp:x")), paragraph_)
        image = "foobar.png"
        width, height, inline = 1111, 2222, element("wp:inline{id=42}")
        document_part_.new_pic_inline.return_value = inline
        InlineShape_.return_value = picture_

        picture = run.add_picture(image, width, height)

        document_part_.new_pic_inline.assert_called_once_with(image, width, height)
        assert run._r.xml == xml("w:r/(wp:x,w:drawing/wp:inline{id=42})")
        InlineShape_.assert_called_once_with(inline)
        assert picture is picture_

    @pytest.mark.parametrize(
        ("initial_r_cxml", "expected_cxml"),
        [
            ("w:r", "w:r"),
            ('w:r/w:t"foo"', "w:r"),
            ("w:r/w:br", "w:r"),
            ("w:r/w:rPr", "w:r/w:rPr"),
            ('w:r/(w:rPr, w:t"foo")', "w:r/w:rPr"),
            (
                'w:r/(w:rPr/(w:b, w:i), w:t"foo", w:cr, w:t"bar")',
                "w:r/w:rPr/(w:b, w:i)",
            ),
        ],
    )
    def it_can_remove_its_content_but_keep_formatting(
        self, initial_r_cxml: str, expected_cxml: str, paragraph_: Mock
    ):
        run = Run(cast(CT_R, element(initial_r_cxml)), paragraph_)

        cleared_run = run.clear()

        assert run._r.xml == xml(expected_cxml)
        assert cleared_run is run

    @pytest.mark.parametrize(
        ("r_cxml", "expected_text"),
        [
            ("w:r", ""),
            ('w:r/w:t"foobar"', "foobar"),
            ('w:r/(w:t"abc", w:tab, w:t"def", w:cr)', "abc\tdef\n"),
            ('w:r/(w:br{w:type=page}, w:t"abc", w:t"def", w:tab)', "abcdef\t"),
        ],
    )
    def it_knows_the_text_it_contains(self, r_cxml: str, expected_text: str, paragraph_: Mock):
        run = Run(cast(CT_R, element(r_cxml)), paragraph_)
        assert run.text == expected_text

    @pytest.mark.parametrize(
        ("new_text", "expected_cxml"),
        [
            ("abc  def", 'w:r/w:t"abc  def"'),
            ("abc\tdef", 'w:r/(w:t"abc", w:tab, w:t"def")'),
            ("abc\ndef", 'w:r/(w:t"abc", w:br,  w:t"def")'),
            ("abc\rdef", 'w:r/(w:t"abc", w:br,  w:t"def")'),
        ],
    )
    def it_can_replace_the_text_it_contains(
        self, new_text: str, expected_cxml: str, paragraph_: Mock
    ):
        run = Run(cast(CT_R, element('w:r/w:t"should get deleted"')), paragraph_)

        run.text = new_text

        assert run._r.xml == xml(expected_cxml)

    # -- fixtures --------------------------------------------------------------------------------

    @pytest.fixture
    def document_part_(self, request: FixtureRequest):
        return instance_mock(request, DocumentPart)

    @pytest.fixture
    def Font_(self, request: FixtureRequest):
        return class_mock(request, "docx.text.run.Font")

    @pytest.fixture
    def font_(self, request: FixtureRequest):
        return instance_mock(request, Font)

    @pytest.fixture
    def InlineShape_(self, request: FixtureRequest):
        return class_mock(request, "docx.text.run.InlineShape")

    @pytest.fixture
    def paragraph_(self, request: FixtureRequest):
        return instance_mock(request, Paragraph)

    @pytest.fixture
    def part_prop_(self, request: FixtureRequest):
        return property_mock(request, Run, "part")

    @pytest.fixture
    def picture_(self, request: FixtureRequest):
        return instance_mock(request, InlineShape)

    @pytest.fixture
    def Text_(self, request: FixtureRequest):
        return class_mock(request, "docx.text.run._Text")



# pyright: reportPrivateUsage=false

"""Unit test suite for the docx.opc.coreprops module."""

from __future__ import annotations

import datetime as dt
from typing import TYPE_CHECKING, cast

import pytest

from docx.opc.coreprops import CoreProperties
from docx.oxml.parser import parse_xml

if TYPE_CHECKING:
    from docx.oxml.coreprops import CT_CoreProperties


class DescribeCoreProperties:
    """Unit-test suite for `docx.opc.coreprops.CoreProperties` objects."""

    @pytest.mark.parametrize(
        ("prop_name", "expected_value"),
        [
            ("author", "python-docx"),
            ("category", ""),
            ("comments", ""),
            ("content_status", "DRAFT"),
            ("identifier", "GXS 10.2.1ab"),
            ("keywords", "foo bar baz"),
            ("language", "US-EN"),
            ("last_modified_by", "Steve Canny"),
            ("subject", "Spam"),
            ("title", "Word Document"),
            ("version", "1.2.88"),
        ],
    )
    def it_knows_the_string_property_values(
        self, prop_name: str, expected_value: str, core_properties: CoreProperties
    ):
        actual_value = getattr(core_properties, prop_name)
        assert actual_value == expected_value

    @pytest.mark.parametrize(
        ("prop_name", "tagname", "value"),
        [
            ("author", "dc:creator", "scanny"),
            ("category", "cp:category", "silly stories"),
            ("comments", "dc:description", "Bar foo to you"),
            ("content_status", "cp:contentStatus", "FINAL"),
            ("identifier", "dc:identifier", "GT 5.2.xab"),
            ("keywords", "cp:keywords", "dog cat moo"),
            ("language", "dc:language", "GB-EN"),
            ("last_modified_by", "cp:lastModifiedBy", "Billy Bob"),
            ("subject", "dc:subject", "Eggs"),
            ("title", "dc:title", "Dissertation"),
            ("version", "cp:version", "81.2.8"),
        ],
    )
    def it_can_change_the_string_property_values(self, prop_name: str, tagname: str, value: str):
        coreProperties = self.coreProperties(tagname="", str_val="")
        core_properties = CoreProperties(cast("CT_CoreProperties", parse_xml(coreProperties)))

        setattr(core_properties, prop_name, value)

        assert core_properties._element.xml == self.coreProperties(tagname, value)

    @pytest.mark.parametrize(
        ("prop_name", "expected_datetime"),
        [
            ("created", dt.datetime(2012, 11, 17, 16, 37, 40, tzinfo=dt.timezone.utc)),
            ("last_printed", dt.datetime(2014, 6, 4, 4, 28, tzinfo=dt.timezone.utc)),
            ("modified", None),
        ],
    )
    def it_knows_the_date_property_values(
        self, prop_name: str, expected_datetime: dt.datetime, core_properties: CoreProperties
    ):
        actual_datetime = getattr(core_properties, prop_name)
        assert actual_datetime == expected_datetime

    @pytest.mark.parametrize(
        ("prop_name", "tagname", "value", "str_val", "attrs"),
        [
            (
                "created",
                "dcterms:created",
                dt.datetime(2001, 2, 3, 4, 5),
                "2001-02-03T04:05:00Z",
                ' xsi:type="dcterms:W3CDTF"',
            ),
            (
                "last_printed",
                "cp:lastPrinted",
                dt.datetime(2014, 6, 4, 4),
                "2014-06-04T04:00:00Z",
                "",
            ),
            (
                "modified",
                "dcterms:modified",
                dt.datetime(2005, 4, 3, 2, 1),
                "2005-04-03T02:01:00Z",
                ' xsi:type="dcterms:W3CDTF"',
            ),
        ],
    )
    def it_can_change_the_date_property_values(
        self, prop_name: str, tagname: str, value: dt.datetime, str_val: str, attrs: str
    ):
        coreProperties = self.coreProperties(tagname="", str_val="")
        core_properties = CoreProperties(cast("CT_CoreProperties", parse_xml(coreProperties)))
        expected_xml = self.coreProperties(tagname, str_val, attrs)

        setattr(core_properties, prop_name, value)

        assert core_properties._element.xml == expected_xml

    @pytest.mark.parametrize(
        ("str_val", "expected_value"),
        [("42", 42), (None, 0), ("foobar", 0), ("-17", 0), ("32.7", 0)],
    )
    def it_knows_the_revision_number(self, str_val: str | None, expected_value: int):
        tagname, str_val = ("cp:revision", str_val) if str_val else ("", "")
        coreProperties = self.coreProperties(tagname, str_val or "")
        core_properties = CoreProperties(cast("CT_CoreProperties", parse_xml(coreProperties)))

        assert core_properties.revision == expected_value

    @pytest.mark.parametrize(("value", "str_val"), [(42, "42")])
    def it_can_change_the_revision_number(self, value: int, str_val: str):
        coreProperties = self.coreProperties(tagname="", str_val="")
        core_properties = CoreProperties(cast("CT_CoreProperties", parse_xml(coreProperties)))
        expected_xml = self.coreProperties("cp:revision", str_val)

        core_properties.revision = value

        assert core_properties._element.xml == expected_xml

    # fixtures -------------------------------------------------------

    def coreProperties(self, tagname: str, str_val: str, attrs: str = "") -> str:
        tmpl = (
            "<cp:coreProperties"
            ' xmlns:cp="http://schemas.openxmlformats.org/package/2006/metadata/core-properties"'
            ' xmlns:dc="http://purl.org/dc/elements/1.1/"'
            ' xmlns:dcmitype="http://purl.org/dc/dcmitype/"'
            ' xmlns:dcterms="http://purl.org/dc/terms/"'
            ' xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"'
            ">%s</cp:coreProperties>\n"
        )
        if not tagname:
            child_element = ""
        elif not str_val:
            child_element = "\n  <%s%s/>\n" % (tagname, attrs)
        else:
            child_element = "\n  <%s%s>%s</%s>\n" % (tagname, attrs, str_val, tagname)
        return tmpl % child_element

    @pytest.fixture
    def core_properties(self):
        element = cast(
            "CT_CoreProperties",
            parse_xml(
                b"<?xml version='1.0' encoding='UTF-8' standalone='yes'?>"
                b'\n<cp:coreProperties xmlns:cp="http://schemas.openxmlformats.o'
                b'rg/package/2006/metadata/core-properties" xmlns:dc="http://pur'
                b'l.org/dc/elements/1.1/" xmlns:dcmitype="http://purl.org/dc/dcm'
                b'itype/" xmlns:dcterms="http://purl.org/dc/terms/" xmlns:xsi="h'
                b'ttp://www.w3.org/2001/XMLSchema-instance">\n'
                b"  <cp:contentStatus>DRAFT</cp:contentStatus>\n"
                b"  <dc:creator>python-docx</dc:creator>\n"
                b'  <dcterms:created xsi:type="dcterms:W3CDTF">2012-11-17T11:07:'
                b"40-05:30</dcterms:created>\n"
                b"  <dc:description/>\n"
                b"  <dc:identifier>GXS 10.2.1ab</dc:identifier>\n"
                b"  <dc:language>US-EN</dc:language>\n"
                b"  <cp:lastPrinted>2014-06-04T04:28:00Z</cp:lastPrinted>\n"
                b"  <cp:keywords>foo bar baz</cp:keywords>\n"
                b"  <cp:lastModifiedBy>Steve Canny</cp:lastModifiedBy>\n"
                b"  <cp:revision>4</cp:revision>\n"
                b"  <dc:subject>Spam</dc:subject>\n"
                b"  <dc:title>Word Document</dc:title>\n"
                b"  <cp:version>1.2.88</cp:version>\n"
                b"</cp:coreProperties>\n"
            ),
        )
        return CoreProperties(element)



"""Utility functions wrapping the excellent `mock` library."""

from __future__ import annotations

from typing import Any
from unittest.mock import (
    ANY,
    MagicMock,
    Mock,
    PropertyMock,
    call,
    create_autospec,
    mock_open,
    patch,
)

from pytest import FixtureRequest, LogCaptureFixture  # noqa: PT013

__all__ = (
    "ANY",
    "FixtureRequest",
    "LogCaptureFixture",
    "MagicMock",
    "Mock",
    "call",
    "class_mock",
    "function_mock",
    "initializer_mock",
    "instance_mock",
    "method_mock",
    "property_mock",
)


def class_mock(
    request: FixtureRequest, q_class_name: str, autospec: bool = True, **kwargs: Any
) -> Mock:
    """Return mock patching class with qualified name `q_class_name`.

    The mock is autospec'ed based on the patched class unless the optional
    argument `autospec` is set to False. Any other keyword arguments are
    passed through to Mock(). Patch is reversed after calling test returns.
    """
    _patch = patch(q_class_name, autospec=autospec, **kwargs)
    request.addfinalizer(_patch.stop)
    return _patch.start()


def cls_attr_mock(
    request: FixtureRequest,
    cls: type,
    attr_name: str,
    name: str | None = None,
    **kwargs: Any,
):
    """Return a mock for attribute `attr_name` on `cls`.

    Patch is reversed after pytest uses it.
    """
    name = request.fixturename if name is None else name
    _patch = patch.object(cls, attr_name, name=name, **kwargs)
    request.addfinalizer(_patch.stop)
    return _patch.start()


def function_mock(
    request: FixtureRequest, q_function_name: str, autospec: bool = True, **kwargs: Any
):
    """Return mock patching function with qualified name `q_function_name`.

    Patch is reversed after calling test returns.
    """
    _patch = patch(q_function_name, autospec=autospec, **kwargs)
    request.addfinalizer(_patch.stop)
    return _patch.start()


def initializer_mock(request: FixtureRequest, cls: type, autospec: bool = True, **kwargs: Any):
    """Return mock for __init__() method on `cls`.

    The patch is reversed after pytest uses it.
    """
    _patch = patch.object(cls, "__init__", autospec=autospec, return_value=None, **kwargs)
    request.addfinalizer(_patch.stop)
    return _patch.start()


def instance_mock(
    request: FixtureRequest,
    cls: type,
    name: str | None = None,
    spec_set: bool = True,
    **kwargs: Any,
):
    """
    Return a mock for an instance of `cls` that draws its spec from the class
    and does not allow new attributes to be set on the instance. If `name` is
    missing or |None|, the name of the returned |Mock| instance is set to
    *request.fixturename*. Additional keyword arguments are passed through to
    the Mock() call that creates the mock.
    """
    name = name if name is not None else request.fixturename
    return create_autospec(cls, _name=name, spec_set=spec_set, instance=True, **kwargs)


def loose_mock(request: FixtureRequest, name: str | None = None, **kwargs: Any):
    """
    Return a "loose" mock, meaning it has no spec to constrain calls on it.
    Additional keyword arguments are passed through to Mock(). If called
    without a name, it is assigned the name of the fixture.
    """
    if name is None:
        name = request.fixturename
    return Mock(name=name, **kwargs)


def method_mock(
    request: FixtureRequest,
    cls: type,
    method_name: str,
    autospec: bool = True,
    **kwargs: Any,
):
    """Return mock for method `method_name` on `cls`.

    The patch is reversed after pytest uses it.
    """
    _patch = patch.object(cls, method_name, autospec=autospec, **kwargs)
    request.addfinalizer(_patch.stop)
    return _patch.start()


def open_mock(request: FixtureRequest, module_name: str, **kwargs: Any):
    """
    Return a mock for the builtin `open()` method in `module_name`.
    """
    target = "%s.open" % module_name
    _patch = patch(target, mock_open(), create=True, **kwargs)
    request.addfinalizer(_patch.stop)
    return _patch.start()


def property_mock(request: FixtureRequest, cls: type, prop_name: str, **kwargs: Any):
    """
    Return a mock for property `prop_name` on class `cls` where the patch is
    reversed after pytest uses it.
    """
    _patch = patch.object(cls, prop_name, new_callable=PropertyMock, **kwargs)
    request.addfinalizer(_patch.stop)
    return _patch.start()


def var_mock(request: FixtureRequest, q_var_name: str, **kwargs: Any):
    """
    Return a mock patching the variable with qualified name `q_var_name`.
    Patch is reversed after calling test returns.
    """
    _patch = patch(q_var_name, **kwargs)
    request.addfinalizer(_patch.stop)
    return _patch.start()



"""Test suite for docx.enum module, focused on base classes.

Configured a little differently because of the meta-programming, the two enumeration
classes at the top constitute the entire fixture and the tests themselves just make
assertions on those.
"""

import enum

import pytest

from docx.enum.base import BaseXmlEnum


class SomeXmlAttr(BaseXmlEnum):
    """SomeXmlAttr docstring."""

    FOO = (1, "foo", "Do foo instead of bar.")
    """Do foo instead of bar."""

    BAR = (2, "bar", "Do bar instead of foo.")
    """Do bar instead of foo."""

    BAZ = (3, None, "Maps to the value assumed when the attribute is omitted.")
    """Maps to the value assumed when the attribute is omitted."""


class DescribeBaseXmlEnum:
    """Unit-test suite for `docx.enum.base.BaseXmlEnum`."""

    def it_is_an_instance_of_EnumMeta_just_like_a_regular_Enum(self):
        assert type(SomeXmlAttr) is enum.EnumMeta

    def it_has_the_same_repr_as_a_regular_Enum(self):
        assert repr(SomeXmlAttr) == "<enum 'SomeXmlAttr'>"

    def it_has_an_MRO_that_goes_through_the_base_class_int_and_Enum(self):
        assert SomeXmlAttr.__mro__ == (
            SomeXmlAttr,
            BaseXmlEnum,
            int,
            enum.Enum,
            object,
        ), f"got: {SomeXmlAttr.__mro__}"

    def it_knows_the_XML_value_for_each_member_by_the_member_instance(self):
        assert SomeXmlAttr.to_xml(SomeXmlAttr.FOO) == "foo"

    def it_knows_the_XML_value_for_each_member_by_the_member_value(self):
        assert SomeXmlAttr.to_xml(2) == "bar"

    def but_it_raises_when_there_is_no_such_member(self):
        with pytest.raises(ValueError, match="42 is not a valid SomeXmlAttr"):
            SomeXmlAttr.to_xml(42)

    def it_can_find_the_member_from_the_XML_attr_value(self):
        assert SomeXmlAttr.from_xml("bar") == SomeXmlAttr.BAR

    def and_it_can_find_the_member_from_None_when_a_member_maps_that(self):
        assert SomeXmlAttr.from_xml(None) == SomeXmlAttr.BAZ

    def but_it_raises_when_there_is_no_such_mapped_XML_value(self):
        with pytest.raises(ValueError, match="SomeXmlAttr has no XML mapping for 'baz'"):
            SomeXmlAttr.from_xml("baz")


class DescribeBaseXmlEnumMembers:
    """Unit-test suite for `docx.enum.base.BaseXmlEnum`."""

    def it_is_an_instance_of_its_XmlEnum_subtype_class(self):
        assert type(SomeXmlAttr.FOO) is SomeXmlAttr

    def it_has_the_default_Enum_repr(self):
        assert repr(SomeXmlAttr.BAR) == "<SomeXmlAttr.BAR: 2>"

    def but_its_str_value_is_customized(self):
        assert str(SomeXmlAttr.FOO) == "FOO (1)"

    def its_value_is_the_same_int_as_its_corresponding_MS_API_enum_member(self):
        assert SomeXmlAttr.FOO.value == 1

    def its_name_is_its_member_name_the_same_as_a_regular_Enum(self):
        assert SomeXmlAttr.FOO.name == "FOO"

    def it_has_an_individual_member_specific_docstring(self):
        assert SomeXmlAttr.FOO.__doc__ == "Do foo instead of bar."

    def it_is_equivalent_to_its_int_value(self):
        assert SomeXmlAttr.FOO == 1
        assert SomeXmlAttr.FOO != 2
        assert SomeXmlAttr.BAR == 2
        assert SomeXmlAttr.BAR != 1



# pyright: reportPrivateUsage=false

"""Unit test suite for the docx.section module."""

from __future__ import annotations

from typing import cast

import pytest

from docx import Document
from docx.enum.section import WD_HEADER_FOOTER, WD_ORIENTATION, WD_SECTION
from docx.oxml.document import CT_Document
from docx.oxml.section import CT_SectPr
from docx.parts.document import DocumentPart
from docx.parts.hdrftr import FooterPart, HeaderPart
from docx.section import Section, Sections, _BaseHeaderFooter, _Footer, _Header
from docx.shared import Inches, Length
from docx.table import Table
from docx.text.paragraph import Paragraph

from .unitutil.cxml import element, xml
from .unitutil.file import test_file
from .unitutil.mock import (
    FixtureRequest,
    Mock,
    call,
    class_mock,
    instance_mock,
    method_mock,
    property_mock,
)


class DescribeSections:
    """Unit-test suite for `docx.section.Sections`."""

    def it_knows_how_many_sections_it_contains(self, document_part_: Mock):
        document_elm = cast(
            CT_Document, element("w:document/w:body/(w:p/w:pPr/w:sectPr, w:sectPr)")
        )
        sections = Sections(document_elm, document_part_)
        assert len(sections) == 2

    def it_can_iterate_over_its_Section_instances(
        self, Section_: Mock, section_: Mock, document_part_: Mock
    ):
        document_elm = cast(
            CT_Document, element("w:document/w:body/(w:p/w:pPr/w:sectPr, w:sectPr)")
        )
        sectPrs = document_elm.xpath("//w:sectPr")
        Section_.return_value = section_
        sections = Sections(document_elm, document_part_)

        section_lst = list(sections)

        assert Section_.call_args_list == [
            call(sectPrs[0], document_part_),
            call(sectPrs[1], document_part_),
        ]
        assert section_lst == [section_, section_]

    def it_can_access_its_Section_instances_by_index(
        self, Section_: Mock, section_: Mock, document_part_: Mock
    ):
        document_elm = cast(
            CT_Document,
            element("w:document/w:body/(w:p/w:pPr/w:sectPr,w:p/w:pPr/w:sectPr,w:sectPr)"),
        )
        sectPrs = document_elm.xpath("//w:sectPr")
        Section_.return_value = section_
        sections = Sections(document_elm, document_part_)

        section_lst = [sections[idx] for idx in range(3)]

        assert Section_.call_args_list == [
            call(sectPrs[0], document_part_),
            call(sectPrs[1], document_part_),
            call(sectPrs[2], document_part_),
        ]
        assert section_lst == [section_, section_, section_]

    def it_can_access_its_Section_instances_by_slice(
        self, Section_: Mock, section_: Mock, document_part_: Mock
    ):
        document_elm = cast(
            CT_Document,
            element("w:document/w:body/(w:p/w:pPr/w:sectPr,w:p/w:pPr/w:sectPr,w:sectPr)"),
        )
        sectPrs = document_elm.xpath("//w:sectPr")
        Section_.return_value = section_
        sections = Sections(document_elm, document_part_)

        section_lst = sections[1:9]

        assert Section_.call_args_list == [
            call(sectPrs[1], document_part_),
            call(sectPrs[2], document_part_),
        ]
        assert section_lst == [section_, section_]

    # -- fixtures---------------------------------------------------------------------------------

    @pytest.fixture
    def document_part_(self, request: FixtureRequest):
        return instance_mock(request, DocumentPart)

    @pytest.fixture
    def Section_(self, request: FixtureRequest):
        return class_mock(request, "docx.section.Section")

    @pytest.fixture
    def section_(self, request: FixtureRequest):
        return instance_mock(request, Section)


class DescribeSection:
    """Unit-test suite for `docx.section.Section`."""

    @pytest.mark.parametrize(
        ("sectPr_cxml", "expected_value"),
        [
            ("w:sectPr", False),
            ("w:sectPr/w:titlePg", True),
            ("w:sectPr/w:titlePg{w:val=0}", False),
            ("w:sectPr/w:titlePg{w:val=1}", True),
            ("w:sectPr/w:titlePg{w:val=true}", True),
        ],
    )
    def it_knows_when_it_displays_a_distinct_first_page_header(
        self, sectPr_cxml: str, expected_value: bool, document_part_: Mock
    ):
        sectPr = cast(CT_SectPr, element(sectPr_cxml))
        section = Section(sectPr, document_part_)

        different_first_page_header_footer = section.different_first_page_header_footer

        assert different_first_page_header_footer is expected_value

    @pytest.mark.parametrize(
        ("sectPr_cxml", "value", "expected_cxml"),
        [
            ("w:sectPr", True, "w:sectPr/w:titlePg"),
            ("w:sectPr/w:titlePg", False, "w:sectPr"),
            ("w:sectPr/w:titlePg{w:val=1}", True, "w:sectPr/w:titlePg"),
            ("w:sectPr/w:titlePg{w:val=off}", False, "w:sectPr"),
        ],
    )
    def it_can_change_whether_the_document_has_distinct_odd_and_even_headers(
        self, sectPr_cxml: str, value: bool, expected_cxml: str, document_part_: Mock
    ):
        sectPr = cast(CT_SectPr, element(sectPr_cxml))
        expected_xml = xml(expected_cxml)
        section = Section(sectPr, document_part_)

        section.different_first_page_header_footer = value

        assert sectPr.xml == expected_xml

    def it_provides_access_to_its_even_page_footer(
        self, document_part_: Mock, _Footer_: Mock, footer_: Mock
    ):
        sectPr = cast(CT_SectPr, element("w:sectPr"))
        _Footer_.return_value = footer_
        section = Section(sectPr, document_part_)

        footer = section.even_page_footer

        _Footer_.assert_called_once_with(sectPr, document_part_, WD_HEADER_FOOTER.EVEN_PAGE)
        assert footer is footer_

    def it_provides_access_to_its_even_page_header(
        self, document_part_: Mock, _Header_: Mock, header_: Mock
    ):
        sectPr = cast(CT_SectPr, element("w:sectPr"))
        _Header_.return_value = header_
        section = Section(sectPr, document_part_)

        header = section.even_page_header

        _Header_.assert_called_once_with(sectPr, document_part_, WD_HEADER_FOOTER.EVEN_PAGE)
        assert header is header_

    def it_provides_access_to_its_first_page_footer(
        self, document_part_: Mock, _Footer_: Mock, footer_: Mock
    ):
        sectPr = cast(CT_SectPr, element("w:sectPr"))
        _Footer_.return_value = footer_
        section = Section(sectPr, document_part_)

        footer = section.first_page_footer

        _Footer_.assert_called_once_with(sectPr, document_part_, WD_HEADER_FOOTER.FIRST_PAGE)
        assert footer is footer_

    def it_provides_access_to_its_first_page_header(
        self, document_part_: Mock, _Header_: Mock, header_: Mock
    ):
        sectPr = cast(CT_SectPr, element("w:sectPr"))
        _Header_.return_value = header_
        section = Section(sectPr, document_part_)

        header = section.first_page_header

        _Header_.assert_called_once_with(sectPr, document_part_, WD_HEADER_FOOTER.FIRST_PAGE)
        assert header is header_

    def it_provides_access_to_its_default_footer(
        self, document_part_: Mock, _Footer_: Mock, footer_: Mock
    ):
        sectPr = cast(CT_SectPr, element("w:sectPr"))
        _Footer_.return_value = footer_
        section = Section(sectPr, document_part_)

        footer = section.footer

        _Footer_.assert_called_once_with(sectPr, document_part_, WD_HEADER_FOOTER.PRIMARY)
        assert footer is footer_

    def it_provides_access_to_its_default_header(
        self, document_part_: Mock, _Header_: Mock, header_: Mock
    ):
        sectPr = cast(CT_SectPr, element("w:sectPr"))
        _Header_.return_value = header_
        section = Section(sectPr, document_part_)

        header = section.header

        _Header_.assert_called_once_with(sectPr, document_part_, WD_HEADER_FOOTER.PRIMARY)
        assert header is header_

    def it_can_iterate_its_inner_content(self):
        document = Document(test_file("sct-inner-content.docx"))

        assert len(document.sections) == 3

        inner_content = list(document.sections[0].iter_inner_content())

        assert len(inner_content) == 3
        p = inner_content[0]
        assert isinstance(p, Paragraph)
        assert p.text == "P1"
        t = inner_content[1]
        assert isinstance(t, Table)
        assert t.rows[0].cells[0].text == "T2"
        p = inner_content[2]
        assert isinstance(p, Paragraph)
        assert p.text == "P3"

        inner_content = list(document.sections[1].iter_inner_content())

        assert len(inner_content) == 3
        t = inner_content[0]
        assert isinstance(t, Table)
        assert t.rows[0].cells[0].text == "T4"
        p = inner_content[1]
        assert isinstance(p, Paragraph)
        assert p.text == "P5"
        p = inner_content[2]
        assert isinstance(p, Paragraph)
        assert p.text == "P6"

        inner_content = list(document.sections[2].iter_inner_content())

        assert len(inner_content) == 3
        p = inner_content[0]
        assert isinstance(p, Paragraph)
        assert p.text == "P7"
        p = inner_content[1]
        assert isinstance(p, Paragraph)
        assert p.text == "P8"
        p = inner_content[2]
        assert isinstance(p, Paragraph)
        assert p.text == "P9"

    @pytest.mark.parametrize(
        ("sectPr_cxml", "expected_value"),
        [
            ("w:sectPr", WD_SECTION.NEW_PAGE),
            ("w:sectPr/w:type", WD_SECTION.NEW_PAGE),
            ("w:sectPr/w:type{w:val=continuous}", WD_SECTION.CONTINUOUS),
            ("w:sectPr/w:type{w:val=nextPage}", WD_SECTION.NEW_PAGE),
            ("w:sectPr/w:type{w:val=oddPage}", WD_SECTION.ODD_PAGE),
            ("w:sectPr/w:type{w:val=evenPage}", WD_SECTION.EVEN_PAGE),
            ("w:sectPr/w:type{w:val=nextColumn}", WD_SECTION.NEW_COLUMN),
        ],
    )
    def it_knows_its_start_type(
        self, sectPr_cxml: str, expected_value: WD_SECTION, document_part_: Mock
    ):
        sectPr = cast(CT_SectPr, element(sectPr_cxml))
        section = Section(sectPr, document_part_)

        start_type = section.start_type

        assert start_type is expected_value

    @pytest.mark.parametrize(
        ("sectPr_cxml", "value", "expected_cxml"),
        [
            (
                "w:sectPr/w:type{w:val=oddPage}",
                WD_SECTION.EVEN_PAGE,
                "w:sectPr/w:type{w:val=evenPage}",
            ),
            ("w:sectPr/w:type{w:val=nextPage}", None, "w:sectPr"),
            ("w:sectPr", None, "w:sectPr"),
            ("w:sectPr/w:type{w:val=continuous}", WD_SECTION.NEW_PAGE, "w:sectPr"),
            ("w:sectPr/w:type", WD_SECTION.NEW_PAGE, "w:sectPr"),
            (
                "w:sectPr/w:type",
                WD_SECTION.NEW_COLUMN,
                "w:sectPr/w:type{w:val=nextColumn}",
            ),
        ],
    )
    def it_can_change_its_start_type(
        self,
        sectPr_cxml: str,
        value: WD_SECTION | None,
        expected_cxml: str,
        document_part_: Mock,
    ):
        sectPr = cast(CT_SectPr, element(sectPr_cxml))
        expected_xml = xml(expected_cxml)
        section = Section(sectPr, document_part_)

        section.start_type = value

        assert section._sectPr.xml == expected_xml

    @pytest.mark.parametrize(
        ("sectPr_cxml", "expected_value"),
        [
            ("w:sectPr/w:pgSz{w:w=1440}", Inches(1)),
            ("w:sectPr/w:pgSz", None),
            ("w:sectPr", None),
        ],
    )
    def it_knows_its_page_width(
        self, sectPr_cxml: str, expected_value: Length | None, document_part_: Mock
    ):
        sectPr = cast(CT_SectPr, element(sectPr_cxml))
        section = Section(sectPr, document_part_)

        page_width = section.page_width

        assert page_width == expected_value

    @pytest.mark.parametrize(
        ("value", "expected_cxml"),
        [
            (None, "w:sectPr/w:pgSz"),
            (Inches(4), "w:sectPr/w:pgSz{w:w=5760}"),
        ],
    )
    def it_can_change_its_page_width(
        self,
        value: Length | None,
        expected_cxml: str,
        document_part_: Mock,
    ):
        sectPr = cast(CT_SectPr, element("w:sectPr"))
        expected_xml = xml(expected_cxml)
        section = Section(sectPr, document_part_)

        section.page_width = value

        assert section._sectPr.xml == expected_xml

    @pytest.mark.parametrize(
        ("sectPr_cxml", "expected_value"),
        [
            ("w:sectPr/w:pgSz{w:h=2880}", Inches(2)),
            ("w:sectPr/w:pgSz", None),
            ("w:sectPr", None),
        ],
    )
    def it_knows_its_page_height(
        self, sectPr_cxml: str, expected_value: Length | None, document_part_: Mock
    ):
        sectPr = cast(CT_SectPr, element(sectPr_cxml))
        section = Section(sectPr, document_part_)

        page_height = section.page_height

        assert page_height == expected_value

    @pytest.mark.parametrize(
        ("value", "expected_cxml"),
        [
            (None, "w:sectPr/w:pgSz"),
            (Inches(2), "w:sectPr/w:pgSz{w:h=2880}"),
        ],
    )
    def it_can_change_its_page_height(
        self, value: Length | None, expected_cxml: str, document_part_: Mock
    ):
        sectPr = cast(CT_SectPr, element("w:sectPr"))
        expected_xml = xml(expected_cxml)
        section = Section(sectPr, document_part_)

        section.page_height = value

        assert section._sectPr.xml == expected_xml

    @pytest.mark.parametrize(
        ("sectPr_cxml", "expected_value"),
        [
            ("w:sectPr/w:pgSz{w:orient=landscape}", WD_ORIENTATION.LANDSCAPE),
            ("w:sectPr/w:pgSz{w:orient=portrait}", WD_ORIENTATION.PORTRAIT),
            ("w:sectPr/w:pgSz", WD_ORIENTATION.PORTRAIT),
            ("w:sectPr", WD_ORIENTATION.PORTRAIT),
        ],
    )
    def it_knows_its_page_orientation(
        self, sectPr_cxml: str, expected_value: WD_ORIENTATION, document_part_: Mock
    ):
        sectPr = cast(CT_SectPr, element(sectPr_cxml))
        section = Section(sectPr, document_part_)

        orientation = section.orientation

        assert orientation is expected_value

    @pytest.mark.parametrize(
        ("value", "expected_cxml"),
        [
            (WD_ORIENTATION.LANDSCAPE, "w:sectPr/w:pgSz{w:orient=landscape}"),
            (WD_ORIENTATION.PORTRAIT, "w:sectPr/w:pgSz"),
            (None, "w:sectPr/w:pgSz"),
        ],
    )
    def it_can_change_its_orientation(
        self, value: WD_ORIENTATION | None, expected_cxml: str, document_part_: Mock
    ):
        sectPr = cast(CT_SectPr, element("w:sectPr"))
        expected_xml = xml(expected_cxml)
        section = Section(sectPr, document_part_)

        section.orientation = value

        assert section._sectPr.xml == expected_xml

    @pytest.mark.parametrize(
        ("sectPr_cxml", "margin_prop_name", "expected_value"),
        [
            ("w:sectPr/w:pgMar{w:left=120}", "left_margin", 76200),
            ("w:sectPr/w:pgMar{w:right=240}", "right_margin", 152400),
            ("w:sectPr/w:pgMar{w:top=-360}", "top_margin", -228600),
            ("w:sectPr/w:pgMar{w:bottom=480}", "bottom_margin", 304800),
            ("w:sectPr/w:pgMar{w:gutter=600}", "gutter", 381000),
            ("w:sectPr/w:pgMar{w:header=720}", "header_distance", 457200),
            ("w:sectPr/w:pgMar{w:footer=840}", "footer_distance", 533400),
            ("w:sectPr/w:pgMar", "left_margin", None),
            ("w:sectPr", "top_margin", None),
        ],
    )
    def it_knows_its_page_margins(
        self,
        sectPr_cxml: str,
        margin_prop_name: str,
        expected_value: int | None,
        document_part_: Mock,
    ):
        sectPr = cast(CT_SectPr, element(sectPr_cxml))
        section = Section(sectPr, document_part_)

        value = getattr(section, margin_prop_name)

        assert value == expected_value

    @pytest.mark.parametrize(
        ("sectPr_cxml", "margin_prop_name", "value", "expected_cxml"),
        [
            ("w:sectPr", "left_margin", Inches(1), "w:sectPr/w:pgMar{w:left=1440}"),
            ("w:sectPr", "right_margin", Inches(0.5), "w:sectPr/w:pgMar{w:right=720}"),
            ("w:sectPr", "top_margin", Inches(-0.25), "w:sectPr/w:pgMar{w:top=-360}"),
            (
                "w:sectPr",
                "bottom_margin",
                Inches(0.75),
                "w:sectPr/w:pgMar{w:bottom=1080}",
            ),
            ("w:sectPr", "gutter", Inches(0.25), "w:sectPr/w:pgMar{w:gutter=360}"),
            (
                "w:sectPr",
                "header_distance",
                Inches(1.25),
                "w:sectPr/w:pgMar{w:header=1800}",
            ),
            (
                "w:sectPr",
                "footer_distance",
                Inches(1.35),
                "w:sectPr/w:pgMar{w:footer=1944}",
            ),
            ("w:sectPr", "left_margin", None, "w:sectPr/w:pgMar"),
            (
                "w:sectPr/w:pgMar{w:top=-360}",
                "top_margin",
                Inches(0.6),
                "w:sectPr/w:pgMar{w:top=864}",
            ),
        ],
    )
    def it_can_change_its_page_margins(
        self,
        sectPr_cxml: str,
        margin_prop_name: str,
        value: Length | None,
        expected_cxml: str,
        document_part_: Mock,
    ):
        sectPr = cast(CT_SectPr, element(sectPr_cxml))
        expected_xml = xml(expected_cxml)
        section = Section(sectPr, document_part_)

        setattr(section, margin_prop_name, value)

        assert section._sectPr.xml == expected_xml

    # -- fixtures-----------------------------------------------------

    @pytest.fixture
    def document_part_(self, request: FixtureRequest):
        return instance_mock(request, DocumentPart)

    @pytest.fixture
    def _Footer_(self, request: FixtureRequest):
        return class_mock(request, "docx.section._Footer")

    @pytest.fixture
    def footer_(self, request: FixtureRequest):
        return instance_mock(request, _Footer)

    @pytest.fixture
    def _Header_(self, request: FixtureRequest):
        return class_mock(request, "docx.section._Header")

    @pytest.fixture
    def header_(self, request: FixtureRequest):
        return instance_mock(request, _Header)


class Describe_BaseHeaderFooter:
    """Unit-test suite for `docx.section._BaseHeaderFooter`."""

    @pytest.mark.parametrize(("has_definition", "expected_value"), [(False, True), (True, False)])
    def it_knows_when_its_linked_to_the_previous_header_or_footer(
        self,
        has_definition: bool,
        expected_value: bool,
        header: _BaseHeaderFooter,
        _has_definition_prop_: Mock,
    ):
        _has_definition_prop_.return_value = has_definition
        assert header.is_linked_to_previous is expected_value

    @pytest.mark.parametrize(
        ("has_definition", "value", "drop_calls", "add_calls"),
        [
            (False, True, 0, 0),
            (True, False, 0, 0),
            (True, True, 1, 0),
            (False, False, 0, 1),
        ],
    )
    def it_can_change_whether_it_is_linked_to_previous_header_or_footer(
        self,
        has_definition: bool,
        value: bool,
        drop_calls: int,
        add_calls: int,
        header: _BaseHeaderFooter,
        _has_definition_prop_: Mock,
        _drop_definition_: Mock,
        _add_definition_: Mock,
    ):
        _has_definition_prop_.return_value = has_definition

        header.is_linked_to_previous = value

        assert _drop_definition_.call_args_list == [call(header)] * drop_calls
        assert _add_definition_.call_args_list == [call(header)] * add_calls

    def it_provides_access_to_the_header_or_footer_part_for_BlockItemContainer(
        self, header: _BaseHeaderFooter, _get_or_add_definition_: Mock, header_part_: Mock
    ):
        # ---this override fulfills part of the BlockItemContainer subclass interface---
        _get_or_add_definition_.return_value = header_part_

        header_part = header.part

        _get_or_add_definition_.assert_called_once_with(header)
        assert header_part is header_part_

    def it_provides_access_to_the_hdr_or_ftr_element_to_help(
        self, header: _BaseHeaderFooter, _get_or_add_definition_: Mock, header_part_: Mock
    ):
        hdr = element("w:hdr")
        _get_or_add_definition_.return_value = header_part_
        header_part_.element = hdr

        hdr_elm = header._element

        _get_or_add_definition_.assert_called_once_with(header)
        assert hdr_elm is hdr

    def it_gets_the_definition_when_it_has_one(
        self,
        header: _BaseHeaderFooter,
        _has_definition_prop_: Mock,
        _definition_prop_: Mock,
        header_part_: Mock,
    ):
        _has_definition_prop_.return_value = True
        _definition_prop_.return_value = header_part_

        header_part = header._get_or_add_definition()

        assert header_part is header_part_

    def but_it_gets_the_prior_definition_when_it_is_linked(
        self,
        header: _BaseHeaderFooter,
        _has_definition_prop_: Mock,
        _prior_headerfooter_prop_: Mock,
        prior_headerfooter_: Mock,
        header_part_: Mock,
    ):
        _has_definition_prop_.return_value = False
        _prior_headerfooter_prop_.return_value = prior_headerfooter_
        prior_headerfooter_._get_or_add_definition.return_value = header_part_

        header_part = header._get_or_add_definition()

        prior_headerfooter_._get_or_add_definition.assert_called_once_with()
        assert header_part is header_part_

    def and_it_adds_a_definition_when_it_is_linked_and_the_first_section(
        self,
        header: _BaseHeaderFooter,
        _has_definition_prop_: Mock,
        _prior_headerfooter_prop_: Mock,
        _add_definition_: Mock,
        header_part_: Mock,
    ):
        _has_definition_prop_.return_value = False
        _prior_headerfooter_prop_.return_value = None
        _add_definition_.return_value = header_part_

        header_part = header._get_or_add_definition()

        _add_definition_.assert_called_once_with(header)
        assert header_part is header_part_

    # -- fixture -----------------------------------------------------

    @pytest.fixture
    def _add_definition_(self, request: FixtureRequest):
        return method_mock(request, _BaseHeaderFooter, "_add_definition")

    @pytest.fixture
    def _definition_prop_(self, request: FixtureRequest):
        return property_mock(request, _BaseHeaderFooter, "_definition")

    @pytest.fixture
    def document_part_(self, request: FixtureRequest):
        return instance_mock(request, DocumentPart)

    @pytest.fixture
    def _drop_definition_(self, request: FixtureRequest):
        return method_mock(request, _BaseHeaderFooter, "_drop_definition")

    @pytest.fixture
    def _get_or_add_definition_(self, request: FixtureRequest):
        return method_mock(request, _BaseHeaderFooter, "_get_or_add_definition")

    @pytest.fixture
    def _has_definition_prop_(self, request: FixtureRequest):
        return property_mock(request, _BaseHeaderFooter, "_has_definition")

    @pytest.fixture
    def header(self, document_part_: Mock) -> _BaseHeaderFooter:
        sectPr = cast(CT_SectPr, element("w:sectPr"))
        return _BaseHeaderFooter(sectPr, document_part_, WD_HEADER_FOOTER.PRIMARY)

    @pytest.fixture
    def header_part_(self, request: FixtureRequest):
        return instance_mock(request, HeaderPart)

    @pytest.fixture
    def prior_headerfooter_(self, request: FixtureRequest):
        return instance_mock(request, _BaseHeaderFooter)

    @pytest.fixture
    def _prior_headerfooter_prop_(self, request: FixtureRequest):
        return property_mock(request, _BaseHeaderFooter, "_prior_headerfooter")


class Describe_Footer:
    """Unit-test suite for `docx.section._Footer`."""

    def it_can_add_a_footer_part_to_help(self, document_part_: Mock, footer_part_: Mock):
        sectPr = cast(CT_SectPr, element("w:sectPr{r:a=b}"))
        document_part_.add_footer_part.return_value = footer_part_, "rId3"
        footer = _Footer(sectPr, document_part_, WD_HEADER_FOOTER.PRIMARY)

        footer_part = footer._add_definition()

        document_part_.add_footer_part.assert_called_once_with()
        assert sectPr.xml == xml("w:sectPr{r:a=b}/w:footerReference{w:type=default,r:id=rId3}")
        assert footer_part is footer_part_

    def it_provides_access_to_its_footer_part_to_help(
        self, document_part_: Mock, footer_part_: Mock
    ):
        sectPr = cast(CT_SectPr, element("w:sectPr/w:footerReference{w:type=even,r:id=rId3}"))
        document_part_.footer_part.return_value = footer_part_
        footer = _Footer(sectPr, document_part_, WD_HEADER_FOOTER.EVEN_PAGE)

        footer_part = footer._definition

        document_part_.footer_part.assert_called_once_with("rId3")
        assert footer_part is footer_part_

    def it_can_drop_the_related_footer_part_to_help(self, document_part_: Mock):
        sectPr = cast(
            CT_SectPr, element("w:sectPr{r:a=b}/w:footerReference{w:type=first,r:id=rId42}")
        )
        footer = _Footer(sectPr, document_part_, WD_HEADER_FOOTER.FIRST_PAGE)

        footer._drop_definition()

        assert sectPr.xml == xml("w:sectPr{r:a=b}")
        document_part_.drop_rel.assert_called_once_with("rId42")

    @pytest.mark.parametrize(
        ("sectPr_cxml", "expected_value"),
        [("w:sectPr", False), ("w:sectPr/w:footerReference{w:type=default}", True)],
    )
    def it_knows_when_it_has_a_definition_to_help(
        self, sectPr_cxml: str, expected_value: bool, document_part_: Mock
    ):
        sectPr = cast(CT_SectPr, element(sectPr_cxml))
        footer = _Footer(sectPr, document_part_, WD_HEADER_FOOTER.PRIMARY)

        has_definition = footer._has_definition

        assert has_definition is expected_value

    def it_provides_access_to_the_prior_Footer_to_help(
        self, request: FixtureRequest, document_part_: Mock, footer_: Mock
    ):
        doc_elm = element("w:document/(w:sectPr,w:sectPr)")
        prior_sectPr, sectPr = cast(CT_SectPr, doc_elm[0]), cast(CT_SectPr, doc_elm[1])
        footer = _Footer(sectPr, document_part_, WD_HEADER_FOOTER.EVEN_PAGE)
        # ---mock must occur after construction of "real" footer---
        _Footer_ = class_mock(request, "docx.section._Footer", return_value=footer_)

        prior_footer = footer._prior_headerfooter

        _Footer_.assert_called_once_with(prior_sectPr, document_part_, WD_HEADER_FOOTER.EVEN_PAGE)
        assert prior_footer is footer_

    def but_it_returns_None_when_its_the_first_footer(self, document_part_: Mock):
        doc_elm = cast(CT_Document, element("w:document/w:sectPr"))
        sectPr = cast(CT_SectPr, doc_elm[0])
        footer = _Footer(sectPr, document_part_, WD_HEADER_FOOTER.PRIMARY)

        prior_footer = footer._prior_headerfooter

        assert prior_footer is None

    # -- fixtures---------------------------------------------------------------------------------

    @pytest.fixture
    def document_part_(self, request: FixtureRequest):
        return instance_mock(request, DocumentPart)

    @pytest.fixture
    def footer_(self, request: FixtureRequest):
        return instance_mock(request, _Footer)

    @pytest.fixture
    def footer_part_(self, request: FixtureRequest):
        return instance_mock(request, FooterPart)


class Describe_Header:
    """Unit-test suite for `docx.section._Header`."""

    def it_can_add_a_header_part_to_help(self, document_part_: Mock, header_part_: Mock):
        sectPr = cast(CT_SectPr, element("w:sectPr{r:a=b}"))
        document_part_.add_header_part.return_value = header_part_, "rId3"
        header = _Header(sectPr, document_part_, WD_HEADER_FOOTER.FIRST_PAGE)

        header_part = header._add_definition()

        document_part_.add_header_part.assert_called_once_with()
        assert sectPr.xml == xml("w:sectPr{r:a=b}/w:headerReference{w:type=first,r:id=rId3}")
        assert header_part is header_part_

    def it_provides_access_to_its_header_part_to_help(
        self, document_part_: Mock, header_part_: Mock
    ):
        sectPr = cast(CT_SectPr, element("w:sectPr/w:headerReference{w:type=default,r:id=rId8}"))
        document_part_.header_part.return_value = header_part_
        header = _Header(sectPr, document_part_, WD_HEADER_FOOTER.PRIMARY)

        header_part = header._definition

        document_part_.header_part.assert_called_once_with("rId8")
        assert header_part is header_part_

    def it_can_drop_the_related_header_part_to_help(self, document_part_: Mock):
        sectPr = cast(
            CT_SectPr, element("w:sectPr{r:a=b}/w:headerReference{w:type=even,r:id=rId42}")
        )
        header = _Header(sectPr, document_part_, WD_HEADER_FOOTER.EVEN_PAGE)

        header._drop_definition()

        assert sectPr.xml == xml("w:sectPr{r:a=b}")
        document_part_.drop_header_part.assert_called_once_with("rId42")

    @pytest.mark.parametrize(
        ("sectPr_cxml", "expected_value"),
        [("w:sectPr", False), ("w:sectPr/w:headerReference{w:type=first}", True)],
    )
    def it_knows_when_it_has_a_header_part_to_help(
        self, sectPr_cxml: str, expected_value: bool, document_part_: Mock
    ):
        sectPr = cast(CT_SectPr, element(sectPr_cxml))
        header = _Header(sectPr, document_part_, WD_HEADER_FOOTER.FIRST_PAGE)

        has_definition = header._has_definition

        assert has_definition is expected_value

    def it_provides_access_to_the_prior_Header_to_help(
        self, request: FixtureRequest, document_part_: Mock, header_: Mock
    ):
        doc_elm = element("w:document/(w:sectPr,w:sectPr)")
        prior_sectPr, sectPr = cast(CT_SectPr, doc_elm[0]), cast(CT_SectPr, doc_elm[1])
        header = _Header(sectPr, document_part_, WD_HEADER_FOOTER.PRIMARY)
        # ---mock must occur after construction of "real" header---
        _Header_ = class_mock(request, "docx.section._Header", return_value=header_)

        prior_header = header._prior_headerfooter

        _Header_.assert_called_once_with(prior_sectPr, document_part_, WD_HEADER_FOOTER.PRIMARY)
        assert prior_header is header_

    def but_it_returns_None_when_its_the_first_header(self, document_part_: Mock):
        doc_elm = element("w:document/w:sectPr")
        sectPr = cast(CT_SectPr, doc_elm[0])
        header = _Header(sectPr, document_part_, WD_HEADER_FOOTER.PRIMARY)

        prior_header = header._prior_headerfooter

        assert prior_header is None

    # -- fixtures---------------------------------------------------------------------------------

    @pytest.fixture
    def document_part_(self, request: FixtureRequest):
        return instance_mock(request, DocumentPart)

    @pytest.fixture
    def header_(self, request: FixtureRequest):
        return instance_mock(request, _Header)

    @pytest.fixture
    def header_part_(self, request: FixtureRequest):
        return instance_mock(request, HeaderPart)



# pyright: reportPrivateUsage=false

"""Unit test suite for the docx.parts.document module."""

import pytest

from docx.comments import Comments
from docx.enum.style import WD_STYLE_TYPE
from docx.opc.constants import CONTENT_TYPE as CT
from docx.opc.constants import RELATIONSHIP_TYPE as RT
from docx.opc.coreprops import CoreProperties
from docx.opc.packuri import PackURI
from docx.package import Package
from docx.parts.comments import CommentsPart
from docx.parts.document import DocumentPart
from docx.parts.hdrftr import FooterPart, HeaderPart
from docx.parts.numbering import NumberingPart
from docx.parts.settings import SettingsPart
from docx.parts.styles import StylesPart
from docx.settings import Settings
from docx.styles.style import BaseStyle
from docx.styles.styles import Styles

from ..unitutil.cxml import element
from ..unitutil.mock import (
    FixtureRequest,
    Mock,
    class_mock,
    instance_mock,
    method_mock,
    property_mock,
)


class DescribeDocumentPart:
    def it_can_add_a_footer_part(
        self, package_: Mock, FooterPart_: Mock, footer_part_: Mock, relate_to_: Mock
    ):
        FooterPart_.new.return_value = footer_part_
        relate_to_.return_value = "rId12"
        document_part = DocumentPart(
            PackURI("/word/document.xml"), CT.WML_DOCUMENT, element("w:document"), package_
        )

        footer_part, rId = document_part.add_footer_part()

        FooterPart_.new.assert_called_once_with(package_)
        relate_to_.assert_called_once_with(document_part, footer_part_, RT.FOOTER)
        assert footer_part is footer_part_
        assert rId == "rId12"

    def it_can_add_a_header_part(
        self, package_: Mock, HeaderPart_: Mock, header_part_: Mock, relate_to_: Mock
    ):
        HeaderPart_.new.return_value = header_part_
        relate_to_.return_value = "rId7"
        document_part = DocumentPart(
            PackURI("/word/document.xml"), CT.WML_DOCUMENT, element("w:document"), package_
        )

        header_part, rId = document_part.add_header_part()

        HeaderPart_.new.assert_called_once_with(package_)
        relate_to_.assert_called_once_with(document_part, header_part_, RT.HEADER)
        assert header_part is header_part_
        assert rId == "rId7"

    def it_can_drop_a_specified_header_part(self, drop_rel_: Mock, package_: Mock):
        document_part = DocumentPart(
            PackURI("/word/document.xml"), CT.WML_DOCUMENT, element("w:document"), package_
        )

        document_part.drop_header_part("rId42")

        drop_rel_.assert_called_once_with(document_part, "rId42")

    def it_provides_access_to_a_footer_part_by_rId(
        self, related_parts_prop_: Mock, related_parts_: Mock, footer_part_: Mock, package_: Mock
    ):
        related_parts_prop_.return_value = related_parts_
        related_parts_.__getitem__.return_value = footer_part_
        document_part = DocumentPart(
            PackURI("/word/document.xml"), CT.WML_DOCUMENT, element("w:document"), package_
        )

        footer_part = document_part.footer_part("rId9")

        related_parts_.__getitem__.assert_called_once_with("rId9")
        assert footer_part is footer_part_

    def it_provides_access_to_a_header_part_by_rId(
        self, related_parts_prop_: Mock, related_parts_: Mock, header_part_: Mock, package_: Mock
    ):
        related_parts_prop_.return_value = related_parts_
        related_parts_.__getitem__.return_value = header_part_
        document_part = DocumentPart(
            PackURI("/word/document.xml"), CT.WML_DOCUMENT, element("w:document"), package_
        )

        header_part = document_part.header_part("rId11")

        related_parts_.__getitem__.assert_called_once_with("rId11")
        assert header_part is header_part_

    def it_can_save_the_package_to_a_file(self, package_: Mock):
        document_part = DocumentPart(
            PackURI("/word/document.xml"), CT.WML_DOCUMENT, element("w:document"), package_
        )

        document_part.save("foobar.docx")

        package_.save.assert_called_once_with("foobar.docx")

    def it_provides_access_to_the_comments_added_to_the_document(
        self, _comments_part_prop_: Mock, comments_part_: Mock, comments_: Mock, package_: Mock
    ):
        comments_part_.comments = comments_
        _comments_part_prop_.return_value = comments_part_
        document_part = DocumentPart(
            PackURI("/word/document.xml"), CT.WML_DOCUMENT, element("w:document"), package_
        )

        assert document_part.comments is comments_

    def it_provides_access_to_the_document_settings(
        self, _settings_part_prop_: Mock, settings_part_: Mock, settings_: Mock, package_: Mock
    ):
        settings_part_.settings = settings_
        _settings_part_prop_.return_value = settings_part_
        document_part = DocumentPart(
            PackURI("/word/document.xml"), CT.WML_DOCUMENT, element("w:document"), package_
        )

        assert document_part.settings is settings_

    def it_provides_access_to_the_document_styles(
        self, _styles_part_prop_: Mock, styles_part_: Mock, styles_: Mock, package_: Mock
    ):
        styles_part_.styles = styles_
        _styles_part_prop_.return_value = styles_part_
        document_part = DocumentPart(
            PackURI("/word/document.xml"), CT.WML_DOCUMENT, element("w:document"), package_
        )

        assert document_part.styles is styles_

    def it_provides_access_to_its_core_properties(self, package_: Mock, core_properties_: Mock):
        document_part = DocumentPart(
            PackURI("/word/document.xml"), CT.WML_DOCUMENT, element("w:document"), package_
        )
        package_.core_properties = core_properties_

        assert document_part.core_properties is core_properties_

    def it_provides_access_to_the_inline_shapes_in_the_document(
        self, InlineShapes_: Mock, package_: Mock
    ):
        document_elm = element("w:document/w:body")
        body_elm = document_elm[0]
        document_part = DocumentPart(
            PackURI("/word/document.xml"), CT.WML_DOCUMENT, document_elm, package_
        )

        inline_shapes = document_part.inline_shapes

        InlineShapes_.assert_called_once_with(body_elm, document_part)
        assert inline_shapes is InlineShapes_.return_value

    def it_provides_access_to_the_numbering_part(
        self, part_related_by_: Mock, numbering_part_: Mock, package_: Mock
    ):
        part_related_by_.return_value = numbering_part_
        document_part = DocumentPart(
            PackURI("/word/document.xml"), CT.WML_DOCUMENT, element("w:document"), package_
        )

        numbering_part = document_part.numbering_part

        part_related_by_.assert_called_once_with(document_part, RT.NUMBERING)
        assert numbering_part is numbering_part_

    def and_it_creates_a_numbering_part_if_not_present(
        self,
        part_related_by_: Mock,
        relate_to_: Mock,
        NumberingPart_: Mock,
        numbering_part_: Mock,
        package_: Mock,
    ):
        part_related_by_.side_effect = KeyError
        NumberingPart_.new.return_value = numbering_part_
        document_part = DocumentPart(
            PackURI("/word/document.xml"), CT.WML_DOCUMENT, element("w:document"), package_
        )

        numbering_part = document_part.numbering_part

        NumberingPart_.new.assert_called_once_with()
        relate_to_.assert_called_once_with(document_part, numbering_part_, RT.NUMBERING)
        assert numbering_part is numbering_part_

    def it_can_get_a_style_by_id(
        self, styles_prop_: Mock, styles_: Mock, style_: Mock, package_: Mock
    ):
        styles_prop_.return_value = styles_
        styles_.get_by_id.return_value = style_
        document_part = DocumentPart(
            PackURI("/word/document.xml"), CT.WML_DOCUMENT, element("w:document"), package_
        )

        style = document_part.get_style("BodyText", WD_STYLE_TYPE.PARAGRAPH)

        styles_.get_by_id.assert_called_once_with("BodyText", WD_STYLE_TYPE.PARAGRAPH)
        assert style is style_

    def it_can_get_the_id_of_a_style(
        self, style_: Mock, styles_prop_: Mock, styles_: Mock, package_: Mock
    ):
        styles_prop_.return_value = styles_
        styles_.get_style_id.return_value = "BodyCharacter"
        document_part = DocumentPart(
            PackURI("/word/document.xml"), CT.WML_DOCUMENT, element("w:document"), package_
        )

        style_id = document_part.get_style_id(style_, WD_STYLE_TYPE.CHARACTER)

        styles_.get_style_id.assert_called_once_with(style_, WD_STYLE_TYPE.CHARACTER)
        assert style_id == "BodyCharacter"

    def it_provides_access_to_its_comments_part_to_help(
        self, package_: Mock, part_related_by_: Mock, comments_part_: Mock
    ):
        part_related_by_.return_value = comments_part_
        document_part = DocumentPart(
            PackURI("/word/document.xml"), CT.WML_DOCUMENT, element("w:document"), package_
        )

        comments_part = document_part._comments_part

        part_related_by_.assert_called_once_with(document_part, RT.COMMENTS)
        assert comments_part is comments_part_

    def and_it_creates_a_default_comments_part_if_not_present(
        self,
        package_: Mock,
        part_related_by_: Mock,
        CommentsPart_: Mock,
        comments_part_: Mock,
        relate_to_: Mock,
    ):
        part_related_by_.side_effect = KeyError
        CommentsPart_.default.return_value = comments_part_
        document_part = DocumentPart(
            PackURI("/word/document.xml"), CT.WML_DOCUMENT, element("w:document"), package_
        )

        comments_part = document_part._comments_part

        CommentsPart_.default.assert_called_once_with(package_)
        relate_to_.assert_called_once_with(document_part, comments_part_, RT.COMMENTS)
        assert comments_part is comments_part_

    def it_provides_access_to_its_settings_part_to_help(
        self, part_related_by_: Mock, settings_part_: Mock, package_: Mock
    ):
        part_related_by_.return_value = settings_part_
        document_part = DocumentPart(
            PackURI("/word/document.xml"), CT.WML_DOCUMENT, element("w:document"), package_
        )

        settings_part = document_part._settings_part

        part_related_by_.assert_called_once_with(document_part, RT.SETTINGS)
        assert settings_part is settings_part_

    def and_it_creates_a_default_settings_part_if_not_present(
        self,
        package_: Mock,
        part_related_by_: Mock,
        SettingsPart_: Mock,
        settings_part_: Mock,
        relate_to_: Mock,
    ):
        part_related_by_.side_effect = KeyError
        SettingsPart_.default.return_value = settings_part_
        document_part = DocumentPart(
            PackURI("/word/document.xml"), CT.WML_DOCUMENT, element("w:document"), package_
        )

        settings_part = document_part._settings_part

        SettingsPart_.default.assert_called_once_with(package_)
        relate_to_.assert_called_once_with(document_part, settings_part_, RT.SETTINGS)
        assert settings_part is settings_part_

    def it_provides_access_to_its_styles_part_to_help(
        self, part_related_by_: Mock, styles_part_: Mock, package_: Mock
    ):
        part_related_by_.return_value = styles_part_
        document_part = DocumentPart(
            PackURI("/word/document.xml"), CT.WML_DOCUMENT, element("w:document"), package_
        )

        styles_part = document_part._styles_part

        part_related_by_.assert_called_once_with(document_part, RT.STYLES)
        assert styles_part is styles_part_

    def and_it_creates_a_default_styles_part_if_not_present(
        self,
        package_: Mock,
        part_related_by_: Mock,
        StylesPart_: Mock,
        styles_part_: Mock,
        relate_to_: Mock,
    ):
        part_related_by_.side_effect = KeyError
        StylesPart_.default.return_value = styles_part_
        document_part = DocumentPart(
            PackURI("/word/document.xml"), CT.WML_DOCUMENT, element("w:document"), package_
        )

        styles_part = document_part._styles_part

        StylesPart_.default.assert_called_once_with(package_)
        relate_to_.assert_called_once_with(document_part, styles_part_, RT.STYLES)
        assert styles_part is styles_part_

    # -- fixtures --------------------------------------------------------------------------------

    @pytest.fixture
    def comments_(self, request: FixtureRequest) -> Mock:
        return instance_mock(request, Comments)

    @pytest.fixture
    def CommentsPart_(self, request: FixtureRequest) -> Mock:
        return class_mock(request, "docx.parts.document.CommentsPart")

    @pytest.fixture
    def comments_part_(self, request: FixtureRequest) -> Mock:
        return instance_mock(request, CommentsPart)

    @pytest.fixture
    def _comments_part_prop_(self, request: FixtureRequest) -> Mock:
        return property_mock(request, DocumentPart, "_comments_part")

    @pytest.fixture
    def core_properties_(self, request: FixtureRequest):
        return instance_mock(request, CoreProperties)

    @pytest.fixture
    def drop_rel_(self, request: FixtureRequest):
        return method_mock(request, DocumentPart, "drop_rel", autospec=True)

    @pytest.fixture
    def FooterPart_(self, request: FixtureRequest):
        return class_mock(request, "docx.parts.document.FooterPart")

    @pytest.fixture
    def footer_part_(self, request: FixtureRequest):
        return instance_mock(request, FooterPart)

    @pytest.fixture
    def HeaderPart_(self, request: FixtureRequest):
        return class_mock(request, "docx.parts.document.HeaderPart")

    @pytest.fixture
    def header_part_(self, request: FixtureRequest):
        return instance_mock(request, HeaderPart)

    @pytest.fixture
    def InlineShapes_(self, request: FixtureRequest):
        return class_mock(request, "docx.parts.document.InlineShapes")

    @pytest.fixture
    def NumberingPart_(self, request: FixtureRequest):
        return class_mock(request, "docx.parts.document.NumberingPart")

    @pytest.fixture
    def numbering_part_(self, request: FixtureRequest):
        return instance_mock(request, NumberingPart)

    @pytest.fixture
    def package_(self, request: FixtureRequest):
        return instance_mock(request, Package)

    @pytest.fixture
    def part_related_by_(self, request: FixtureRequest):
        return method_mock(request, DocumentPart, "part_related_by")

    @pytest.fixture
    def relate_to_(self, request: FixtureRequest):
        return method_mock(request, DocumentPart, "relate_to")

    @pytest.fixture
    def related_parts_(self, request: FixtureRequest):
        return instance_mock(request, dict)

    @pytest.fixture
    def related_parts_prop_(self, request: FixtureRequest):
        return property_mock(request, DocumentPart, "related_parts")

    @pytest.fixture
    def SettingsPart_(self, request: FixtureRequest):
        return class_mock(request, "docx.parts.document.SettingsPart")

    @pytest.fixture
    def settings_(self, request: FixtureRequest):
        return instance_mock(request, Settings)

    @pytest.fixture
    def settings_part_(self, request: FixtureRequest):
        return instance_mock(request, SettingsPart)

    @pytest.fixture
    def _settings_part_prop_(self, request: FixtureRequest):
        return property_mock(request, DocumentPart, "_settings_part")

    @pytest.fixture
    def style_(self, request: FixtureRequest):
        return instance_mock(request, BaseStyle)

    @pytest.fixture
    def styles_(self, request: FixtureRequest):
        return instance_mock(request, Styles)

    @pytest.fixture
    def StylesPart_(self, request: FixtureRequest):
        return class_mock(request, "docx.parts.document.StylesPart")

    @pytest.fixture
    def styles_part_(self, request: FixtureRequest):
        return instance_mock(request, StylesPart)

    @pytest.fixture
    def styles_prop_(self, request: FixtureRequest):
        return property_mock(request, DocumentPart, "styles")

    @pytest.fixture
    def _styles_part_prop_(self, request: FixtureRequest):
        return property_mock(request, DocumentPart, "_styles_part")



"""Step implementations for hyperlink-related features."""

from __future__ import annotations

from typing import Dict, Tuple

from behave import given, then
from behave.runner import Context

from docx import Document

from helpers import test_docx

# given ===================================================


@given("a hyperlink")
def given_a_hyperlink(context: Context):
    document = Document(test_docx("par-hyperlinks"))
    context.hyperlink = document.paragraphs[1].hyperlinks[0]


@given("a hyperlink having a URI fragment")
def given_a_hyperlink_having_a_uri_fragment(context: Context):
    document = Document(test_docx("par-hlink-frags"))
    context.hyperlink = document.paragraphs[1].hyperlinks[0]


@given("a hyperlink having address {address} and fragment {fragment}")
def given_a_hyperlink_having_address_and_fragment(context: Context, address: str, fragment: str):
    paragraph_idxs: Dict[Tuple[str, str], int] = {
        ("''", "linkedBookmark"): 1,
        ("https://foo.com", "''"): 2,
        ("https://foo.com?q=bar", "''"): 3,
        ("http://foo.com/", "intro"): 4,
        ("https://foo.com?q=bar#baz", "''"): 5,
        ("court-exif.jpg", "''"): 7,
    }
    paragraph_idx = paragraph_idxs[(address, fragment)]
    document = Document(test_docx("par-hlink-frags"))
    paragraph = document.paragraphs[paragraph_idx]
    context.hyperlink = paragraph.hyperlinks[0]


@given("a hyperlink having {zero_or_more} rendered page breaks")
def given_a_hyperlink_having_rendered_page_breaks(context: Context, zero_or_more: str):
    paragraph_idx = {
        "no": 1,
        "one": 2,
    }[zero_or_more]
    document = Document(test_docx("par-hyperlinks"))
    paragraph = document.paragraphs[paragraph_idx]
    context.hyperlink = paragraph.hyperlinks[0]


@given("a hyperlink having {one_or_more} runs")
def given_a_hyperlink_having_one_or_more_runs(context: Context, one_or_more: str):
    paragraph_idx, hyperlink_idx = {
        "one": (1, 0),
        "two": (2, 1),
    }[one_or_more]
    document = Document(test_docx("par-hyperlinks"))
    paragraph = document.paragraphs[paragraph_idx]
    context.hyperlink = paragraph.hyperlinks[hyperlink_idx]


# then =====================================================


@then("hyperlink.address is the URL of the hyperlink")
def then_hyperlink_address_is_the_URL_of_the_hyperlink(context: Context):
    actual_value = context.hyperlink.address
    expected_value = "http://yahoo.com/"
    assert actual_value == expected_value, f"expected: {expected_value}, got: {actual_value}"


@then("hyperlink.contains_page_break is {value}")
def then_hyperlink_contains_page_break_is_value(context: Context, value: str):
    actual_value = context.hyperlink.contains_page_break
    expected_value = {"True": True, "False": False}[value]
    assert actual_value == expected_value, f"expected: {expected_value}, got: {actual_value}"


@then("hyperlink.fragment is the URI fragment of the hyperlink")
def then_hyperlink_fragment_is_the_URI_fragment_of_the_hyperlink(context: Context):
    actual_value = context.hyperlink.fragment
    expected_value = "linkedBookmark"
    assert actual_value == expected_value, f"expected: {expected_value}, got: {actual_value}"


@then("hyperlink.runs contains only Run instances")
def then_hyperlink_runs_contains_only_Run_instances(context: Context):
    actual_value = [type(item).__name__ for item in context.hyperlink.runs]
    expected_value = ["Run" for _ in context.hyperlink.runs]
    assert actual_value == expected_value, f"expected: {expected_value}, got: {actual_value}"


@then("hyperlink.runs has length {value}")
def then_hyperlink_runs_has_length(context: Context, value: str):
    actual_value = len(context.hyperlink.runs)
    expected_value = int(value)
    assert actual_value == expected_value, f"expected: {expected_value}, got: {actual_value}"


@then("hyperlink.text is the visible text of the hyperlink")
def then_hyperlink_text_is_the_visible_text_of_the_hyperlink(context: Context):
    actual_value = context.hyperlink.text
    expected_value = "awesome hyperlink"
    assert actual_value == expected_value, f"expected: {expected_value}, got: {actual_value}"


@then("hyperlink.url is {value}")
def then_hyperlink_url_is_value(context: Context, value: str):
    actual_value = context.hyperlink.url
    expected_value = "" if value == "''" else value
    assert actual_value == expected_value, f"expected: {expected_value}, got: {actual_value}"



"""Test suite for the docx.oxml.text.hyperlink module."""

from typing import cast

import pytest

from docx.oxml.text.hyperlink import CT_Hyperlink
from docx.oxml.text.run import CT_R

from ...unitutil.cxml import element


class DescribeCT_Hyperlink:
    """Unit-test suite for the CT_Hyperlink (<w:hyperlink>) element."""

    def it_has_a_relationship_that_contains_the_hyperlink_address(self):
        cxml = 'w:hyperlink{r:id=rId6}/w:r/w:t"post"'
        hyperlink = cast(CT_Hyperlink, element(cxml))

        rId = hyperlink.rId

        assert rId == "rId6"

    @pytest.mark.parametrize(
        ("cxml", "expected_value"),
        [
            # -- default (when omitted) is True, somewhat surprisingly --
            ("w:hyperlink{r:id=rId6}", True),
            ("w:hyperlink{r:id=rId6,w:history=0}", False),
            ("w:hyperlink{r:id=rId6,w:history=1}", True),
        ],
    )
    def it_knows_whether_it_has_been_clicked_on_aka_visited(self, cxml: str, expected_value: bool):
        hyperlink = cast(CT_Hyperlink, element(cxml))
        assert hyperlink.history is expected_value

    def it_has_zero_or_more_runs_containing_the_hyperlink_text(self):
        cxml = 'w:hyperlink{r:id=rId6,w:history=1}/(w:r/w:t"blog",w:r/w:t" post")'
        hyperlink = cast(CT_Hyperlink, element(cxml))

        rs = hyperlink.r_lst

        assert [type(r) for r in rs] == [CT_R, CT_R]
        assert rs[0].text == "blog"
        assert rs[1].text == " post"
